/*  pdnmesh - a 2D finite element solver
    Copyright (c) 2001 Sarod Yatawatta <sarod@cs.pdn.ac.lk>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <math.h> /* for sqrt() */
#include "types.h"


/* globals */
unsigned long int npoints;/* number of points in the input, except boundary*/
unsigned long int nedges; 
/* total number of possible edges for used in delaunay */
/* triangulation */

unsigned long int bedges; /* total number of boundary edges */
edge *edgearray; /* edge array - including all boundaries */
unsigned long int nbound; /* number of boundaries */
bnode *boundarray; /* array of boundaries */

elist * frontier = 0; /* list of edges in frontier */
edge hull; /* the first edge to start triangulation */
unsigned long int triangle_count = 0; /* total triangles in triangulation */

long double xoffset; /* x offset max+min/2*/
long double yoffset; /* y offset max+min/2*/
long double xrange; /* x range max - min */
long double yrange; /* y range max - min */
/* using these 4, we can construct the original coordinates */
long double scale; /* scale */



/* global no for current boundary */
unsigned long int boundary_no;

/* function to get an edge from list */
/* __LOCAL__ */
static elist * 
E_get( elist * Lhead, edge * tt )
{
 
  elist * temp;
  if ( Lhead == 0 ) {
	 tt = 0;
	return(0);
  }	/* NULL */
  *tt = Lhead->data;
  temp = Lhead;
  Lhead = Lhead->next;
  free(temp); 
  return(Lhead);
}

/* function to insert an edge to list -- a queue */
static elist 
*E_insert(  elist * Lhead , edge Ldata )
{
  elist *temp, *tail;
 
  temp = (elist *)malloc(sizeof(elist));
  temp->data.p1 = Ldata.p1;
  temp->data.p2 = Ldata.p2;
  /* first the null case */
  if ( Lhead == NULL ) {
    temp->next = Lhead;
    Lhead = temp;
    return( Lhead );
  }

  /* list is not null */ 
  /* find tha tail of the list */
   tail = Lhead;
   while( tail->next )
     tail = tail->next;
   /* now we reach the last element */
   temp->next = NULL;    
   tail->next = temp;
 
   return( Lhead );
}


/* skips comment lines */
static int 
skip_lines(FILE *fin)
{

  char c;
  
  do {
    if ( ( c = getc(fin) ) == EOF )
      return(-1);
    
    /* handle empty lines */
    if ( c == '\n' )
      continue; /* next line */
    if ( (c != '#') ) {
      ungetc(c,fin);
      return(0);
    } else { /* skip this line */
      do { 
	if ( ( c = getc(fin) ) == EOF )
	  return(-1);
      } while (  c != '\n') ;
    }
  } while( 1 );
}

/* skips rest of line */
static int 
skip_restof_line(FILE *fin)
{
  char c;
  
  do { 
    if ( ( c = getc(fin) ) == EOF )
      return(-1);
  } while (  c != '\n') ;
  return(1);
}


/* function to get the expression for rho from input and build parse tree */
int
insert_expression(exp_nodetype *p, unsigned long int boundary)
{
		  boundarray[boundary].rhoexp = p;
		  return(0);
}


/* this function opens a file and reads coordinates */
/* and also adds a bounding bow to enclose the input points */
/* also normailze all coordinates so none will be higher than (1.0,1.0) */
/* this is required for plotting  */
/* __ GLOBAL __ */
static void 
getinput( void )
{
  FILE *infd; /* input file */
  unsigned long int i;
  long double ymin, xmin, ymax, xmax, temp1, temp2 ;
  /* note maximum can be either an x coord or a y coord */
  unsigned long int lowestx,next; /* need one hull edge */
  /* these are the edge points */
  point pp;
  point * p; /* array of all input points */
  /* note this is not the mesh points later introduced */

  long double distance, cur_distance, gradient, cur_gradient;

  /* buffer for reading expression for rho */
  char *buff,c;
  unsigned int buff_len,k; /* size of buffer , index of buffer */
  
  infd =fopen( infile, "r" );
  if ( infd == NULL ) {
    fprintf(stderr, "could not open file %s\n", infile);
    exit(1);
  }

  skip_lines(infd);
  /* read the number of points */
  if ( fscanf(infd, "%ld" , &npoints ) == EOF ) {
    fprintf(stderr, "could not read file %s\n", infile);
    exit(1);
  }
  skip_restof_line(infd);
  
 
 
  init_mesh(&M, npoints);
  maxpoints = npoints; 
  /* remember to update the maximum points limit */
  /* assign very high values to lowest initial values */
  /* then there will always be a value low than this */
  ymin=xmin=MAX;
  ymax=xmax=-1*MAX;
  
 
  pp.pos = BOUNDARY; /* naturally */
  /* now read coords */
  /* read pre-assigned point number as well */
  for ( i = 0; i < npoints; i++) {
    /* read point  number, x, y, potential, fixed or not */
    skip_lines(infd);
    if ( fscanf(infd, "%ld %Lf %Lf %Lf %ld" , &next, &(pp.x), &(pp.y), &(pp.pot), &lowestx ) == EOF ) {
      fprintf(stderr, "could not read file %s\n", infile);
      exit(1);
    }
    skip_restof_line(infd);
    if ( lowestx == 1 ) /* boundary, fixed point */
      pp.val = FX;
    else
      pp.val = VAR;
    if ( next != insert(pp)) {
      fprintf(stderr, "getinput:  point numbers changed at %ld\n file format error\n", next);
      exit(1);
    }
  }

  /* now read the boundary edges from input */
  skip_lines(infd);
  if ( fscanf(infd, "%ld" , &bedges ) == EOF ) {
    fprintf(stderr, "could not read file %s\n", infile);
    exit(1);
  }
  skip_restof_line(infd);
  /* now allocate memory for boundary */
  edgearray =(edge *)malloc((size_t)bedges*sizeof(edge));
    
  /* now read edge coordinates */
  /* read the type of edge as well */
  for ( i = 0; i < bedges; i++ ) {
    /* read edge no., point1, point2, dirichlet/neumann */
     skip_lines(infd);
    if ( fscanf(infd, "%ld %ld %ld %ld" , &lowestx, &(edgearray[i].p1), &(edgearray[i].p2), &next) == EOF ) {
      fprintf(stderr, "could not read file %s\n", infile);
      exit(1);
    }
    skip_restof_line(infd);
    if ( next == 1 )  /* a dirichlet boundary */
      edgearray[i].type = DR;
    else 
      edgearray[i].type = NU; /* neumann boundray */
  }
  
  /* now read the boundaries */
  /* first the number of boundaries */
  skip_lines(infd);
  if ( fscanf(infd, "%ld" , &nbound ) == EOF ) {
    fprintf(stderr, "could not read file %s\n", infile);
    exit(1);
  }
  skip_restof_line(infd);
  /* remember nbound > 1 because we always have outer boundaries */
  /* now allocate memory for the boundary array */
  boundarray = (bnode *)calloc(nbound, sizeof(bnode));
  /* now read boundaries one by one */
  for ( i = 0; i < nbound; i++ ) {
    /* read the no of edges, type, rho, epsilon */
    skip_lines(infd);
    if ( fscanf(infd, "%ld %d %Lf " , &(boundarray[i].edge_count),
		&(boundarray[i].type), &(boundarray[i].epsilon)) == EOF ) {
      fprintf(stderr, "could not read file %s\n", infile);
      exit(1);
    }
    
    /* now read expression only if boundary not deleted */
    
    buff_len = 30; k = 0;
    buff = (char*)malloc(sizeof(char)*(size_t)(buff_len+2));
    while ( ((c = fgetc(infd)) != EOF ) && k < buff_len) {
      buff[k++] = c;
      if ( c == '\n')
	break;
    }
    if ( c != '\n') {
      buff_len += 30;
      buff = (char*)realloc((void*)buff,sizeof(char)*(size_t)(buff_len+2));
      buff[k++] = c;
      while ( ((c = fgetc(infd)) != EOF ) && k < buff_len ) {
	buff[k++] = c;
	if ( c == '\n')
	  break;
      }
    }
    buff[k++] = ';';
    buff[k++] = '\0';
    
    /* make global boundary this -- used in parser */
    boundary_no = i;
    /* now parse equation */
#ifdef DEBUG
	 printf("input: parsing rho for %ld\n",boundary_no);
#endif
	 if ( boundarray[i].type != 0 ) 
	   equation_parse(buff);
	 
	 free(buff);
	 /* by default boundary is closed */
	 
	 /* now get ready to read the edges of the boundary */
	 /* first allocate memory */
	 boundarray[i].edges = (unsigned long*)calloc(boundarray[i].edge_count,
						      sizeof(unsigned long));
	 for ( next = 0; next < boundarray[i].edge_count ; next++ ) {
	   skip_lines(infd);
	   if ( fscanf(infd, "%ld" , &(boundarray[i].edges[next]) ) == EOF ) {
	     fprintf(stderr, "could not read file %s\n", infile);
	     exit(1);
	   }
	   skip_restof_line(infd);
	 }
	   
  }

  /* here split long boundaries */
  for ( i = 0; i < bedges; i++ ) {
    distance = (Mx(edgearray[i].p1)-Mx(edgearray[i].p2))*(Mx(edgearray[i].p1)-Mx(edgearray[i].p2)) 
      +(My(edgearray[i].p1)-My(edgearray[i].p2))*(My(edgearray[i].p1)-My(edgearray[i].p2));

    temp1 = 5*maximum_area*badness_limit; 
	 /* distance for splittiong */
    if ( distance > temp1 ) {
#ifdef DEBUG 
     printf("need to split edge %ld\n", i);
#endif
     /* now add more point inbetween */
     /* first determine how many */
     next = (unsigned long int )sqrt(distance/(300*maximum_area*badness_limit))+1;
     /* at least 2 */
     if ( edgearray[i].type == DR )
	 pp.val = FX;
       else
	 pp.val = VAR;
     pp.pos = BOUNDARY;
     for ( lowestx = 1; lowestx < next; lowestx++ ) {
       pp.x = (lowestx*Mx(edgearray[i].p1)+(next-lowestx)*Mx(edgearray[i].p2))/next;
       pp.y = (lowestx*My(edgearray[i].p1)+(next-lowestx)*My(edgearray[i].p2))/next;
       pp.pot = (lowestx*Mpot(edgearray[i].p1)+(next-lowestx)*Mpot(edgearray[i].p2))/next;
#ifdef DEBUG
       printf("inserting new point\n");
#endif
       insert(pp);
     }
    }
  }
  
  
  /* now check for duplicate points */
  /* then no of points will be lower */
  npoints = M.count;
  p = (point *)calloc(npoints, sizeof(point));
  /* now insert the true points back to array */
  for ( i = 0; i < npoints; i++ ) {
    p[i].x = Mx(i);
    p[i].y = My(i);
#ifdef DEBUG
    printf("%d read %Lf, %Lf %Lf\n", i , p[i].x, p[i].y, p[i].pot); 
#endif    
    p[i].pos = BOUNDARY; /* we take points at the boundary to be inside */
  }

  
  for ( i = 0; i < npoints; i++ ) {
    /* now check the lowest */
    if ( p[i].y < ymin ) 
      ymin = p[i].y;
    if (p[i].x < xmin ) {
      lowestx = i;
      xmin = p[i].x;
    } else if ( p[i].x - xmin == 0.0 ) {
      /* choose the one with lowest y */
      if ( p[i].y < p[lowestx].y ) {
	lowestx = i;
      }
    }
     if ( p[i].y > ymax )
      ymax = p[i].y;
    if (p[i].x > xmax )
      xmax = p[i].x;  
      
  }
#ifdef DEBUG
  printf("lowestx %d\n", lowestx);
#endif
  /* now determine the first edge for frontier, hull edge */
  /* by finding the angle made by all other points with */
  /* "lowestx" point */
  cur_gradient = -10000000000000000000000.0 ; /* very low value */
  cur_distance = 100000000000000000000000.0; /* very high value */
  
  next = 0; /* to be in the safe side for small input */
  /* i.e. 2 or 3 points */

  for ( i = 0; i <npoints; i++ ) {
    if ( i != lowestx ) {
      
#ifdef DEBUG
      printf("i=%d ",i);
#endif      
      /* first calculate gradient */
      temp1 = p[i].x - p[lowestx].x;
      temp2 = p[i].y - p[lowestx].y;
      
#ifdef DEBUG
      printf("temp1= %Lf temp2= %Lf",temp1,temp2);
#endif      
      if ( temp1 == 0.0f ) 
	gradient = 100000000000000000000.0; /* very high positive value */
      else
	gradient = temp2/temp1;
      
      
      /* now compare gradients */
      if ( gradient == cur_gradient ) {
	/* find distance */
	/* just compare the square od distance */
	distance = temp1*temp1 + temp2*temp2;
	
	if ( distance < cur_distance ) {
	  next = i;
	  cur_distance = distance;
	}
      } else if ( gradient > cur_gradient ) {
	next = i;
	cur_gradient = gradient;
	cur_distance = temp1*temp1 + temp2*temp2;
      }
    }
#ifdef DEBUG
  printf("next %d\n",next);
#endif
  }
  
  /* hull edge */
  hull.p2 = lowestx;
  hull.p1 = next;
#ifdef DEBUG
  printf("hull edge %ld %ld\n", hull.p1, hull.p2 );
  printf("from (%Lf, %Lf) to (%Lf, %Lf)\n",p[hull.p1].x,p[hull.p1].y,p[hull.p2].x,p[hull.p2].y);
#endif
  
  xoffset =  -(xmax+xmin)*0.5;
  yoffset =  -(ymax+ymin)*0.5;
  xrange = (xmax-xmin)*0.5;
  yrange = (ymax-ymin)*0.5;
  /* now reorganize coordinates for nice plotting */
  /* inside a box of 2x2 centered at origin */
  for ( i = 0; i < npoints ; i++ ) {	
    p[i].x +=xoffset; 
    p[i].y +=yoffset;
  }
  /* now find the maximum value of coordinates */
  ymax = fabsl(ymax);
  xmax = fabsl(xmax);
  xmin = fabsl(xmin);
  ymin = fabsl(ymin);
  if ( xmax < ymax )
    xmax = ymax;
  if ( xmax < ymin )
    xmax = ymin;
  if ( xmax < xmin  )
    xmax = xmin;
  /* now xmax has ordinate with maximum magnitude */

  /* printf("lowest is %d , %Lf , %Lf\n", lowest, p[lowest].x, p[lowest].y);*/
  /*printf("maximum values is %Lf\n", maximum ); */
  /* allow a margin about 20%, so increase maximum by 1.2 */
  xmax *= 1.2;
  
  scale = xmax; /* global scale */
  
  
  /* now normalize points and insert to mesh */
  xmax = 1/xmax; /* easier to multyply than divide */
 
  for ( i = 0; i < npoints ; i++ ) {	
    p[i].x = p[i].x*xmax;
    Mx(i) = p[i].x;
    p[i].y = p[i].y*xmax;
    My(i) = p[i].y;
  }
 

  /* max possible number of edges */
  nedges = (int)((float)npoints*((float)npoints-1.0)*0.5);
 



  

#ifdef DEBUG
  /* print out a summery */
  for ( i = 0; i < nbound; i++ ) {
    printf("boundary %ld edges %ld epsilon %Lf\n",
	   i, boundarray[i].edge_count, 
	    boundarray[i].epsilon);
    for ( next = 0 ; next < boundarray[i].edge_count ; next++ ) { 
      printf("%ld ", boundarray[i].edges[next]);
      if ( edgearray[boundarray[i].edges[next]].type == DR  )
	printf("dirichlet\n");
      else
	printf("numann\n");
    }
    printf("\n");
  }
#endif

  free(p); 
  fclose( infd );
}




/* function to evaluate the determinant of three points */
/* ___ LOCAL __ */
static long double 
determ( long double x1, long double y1, long double x2, long double y2, long double x3, long double y3 ) {
  return ( x1*(y2-y3)+x2*(y3-y1)+x3*(y1-y2) );
}

/* a function to calculate the area of a triangle */
long double 
area( triag t )
{
  long double temp1;
  temp1 = determ(Mx(t.n1),My(t.n1),Mx(t.n2),My(t.n2),Mx(t.n3),My(t.n3));
  return(0.5*fabs(temp1));
}

/* function to check whether three points are counter clockwise */
/* given the input sequence */

static int 
ccw1( long double p0x,long double p0y, long double p1x,long double p1y, long double p2x,long double p2y)
{
  
#ifdef DEBUG
  printf("determ %Lf\n",determ(p0x, p0y, p1x, p1y, p2x, p2y));
  return( determ(p0x, p0y, p1x, p1y, p2x, p2y) > TOL );
#else
  return( determ(p0x, p0y, p1x, p1y, p2x, p2y) > TOL );
#endif
}


/* to check three points p1,p2, p3 are ccw */
int 
ccw2( unsigned long p1, unsigned long p2, unsigned long p3 )
{
#ifdef DEBUG
  if ( determ(Mx(p1), My(p1), Mx(p2), My(p2), Mx(p3), My(p3)) > TOL ) {
    printf("points %ld %ld %ld are ccw\n", p1,p2,p3);
    return(1);
  } else {
    printf("points %ld %ld %ld are NOT ccw\n", p1,p2,p3);
    return(0);
  }
#endif
  return(determ(Mx(p1), My(p1), Mx(p2), My(p2), Mx(p3), My(p3)) > TOL);
}


/* to check intersection of two lines */
/* but not exact intersection */
/* intersection at edges is not taken */
/* to be valid */
static int 
intersect( edge e1, edge e2 )
{
 long double temp1, temp2;
 temp1 = determ(Mx(e1.p1),My(e1.p1),Mx(e1.p2),My(e1.p2),Mx(e2.p1),My(e2.p1)); 
 temp2 = determ(Mx(e1.p1),My(e1.p1),Mx(e1.p2),My(e1.p2),Mx(e2.p2),My(e2.p2)); 
 // printf("intersect: %Lf %Lf\n",temp1,temp2);
 if ( ( (temp1 > TOL ) && ( temp2 < -TOL ) ) 
      ||( (temp1 < -TOL ) && ( temp2 > TOL ) ) ) {
   /* check with changed points */
   temp1 = determ(Mx(e2.p1),My(e2.p1),Mx(e2.p2),My(e2.p2),Mx(e1.p1),My(e1.p1)); 
   temp2 = determ(Mx(e2.p1),My(e2.p1),Mx(e2.p2),My(e2.p2),Mx(e1.p2),My(e1.p2)); 
   //printf("intersect: %Lf %Lf\n",temp1,temp2);
   if ( ( (temp1 > TOL ) && ( temp2 < -TOL ) ) 
	||( (temp1 < -TOL ) && ( temp2 > TOL ) ) ) {
     //printf("intersect: yes (%ld,%ld) (%ld,%ld)\n",e1.p1,e1.p2,e2.p1,e2.p2);
     return(1);
   }
   return(0);
 }
 return(0); /* no intersection */
}

/* returns the intersecting point if any */
/* if not, return NULL */
static int 
check_intersection( unsigned long int p1, unsigned long int p2, edge boundary[] , point *pt )
{
  edge e;
  
  long double x,y;
  long double x1,x2,x3,x4,y1,y2,y3,y4;
  int i;
  e.p1 = p1;
  e.p2 = p2;
  /* more than one edge may intersect, then we take the */
  /* closest edge to point p1 */
  for ( i = 0; i < bedges; i++ ) 
    if ( intersect( e, boundary[i] ) ){
      /* calculate the intersection */
      /* and calculate the distance from p1 */
#ifdef DEBUG      
      printf("intersection found ");
#endif      
      x1 = Mx(p1);
      y1 = My(p1);
      x2 = Mx(p2);
      y2 = My(p2);
      x3 = Mx(boundary[i].p1);
      y3 = My(boundary[i].p1);
      x4 = Mx(boundary[i].p2);
      y4 = My(boundary[i].p2);
      x = (x1*(y2-y1)*(x4-x3)-x3*(x2-x1)*(y4-y3)+(y3-y1)*(x2-x1)*(x4-x3))/((y2-y1)*(x4-x3)-(y4-y3)*(x2-x1));
      if ( x2 != x1 )
	y = (x-x1)*(y2-y1)/(x2-x1)+y1;
      else /* assume x3 != x4 */
	y = (x-x3)*(y4-y3)/(x4-x3)+y3;
      /* assume (x,y) not already included */
#ifdef DEBUG      
      printf("at %Lf %Lf lines (%ld, %ld), (%ld, %ld)\n",x,y,e.p1,e.p2,boundary[i].p1,boundary[i].p2);
#endif
      pt->x = x;
      pt->y = y;
      /* we take the interpolated value of potential */
      /* between points 3,4 on the boundary */
      if ( fabsl(x4 - x3) > TOL )
	pt->pot = (x-x3)*(Mpot(boundary[i].p2)-Mpot(boundary[i].p1))/(x4-x3)+Mpot(boundary[i].p1);
      else /* interpolate on y coordinate */
	pt->pot = (y-y3)*(Mpot(boundary[i].p2)-Mpot(boundary[i].p1))/(y4-y3)+Mpot(boundary[i].p1);
      //      printf("%Lf\n",pt->pot);
     
      if ( boundary[i].type == DR )
	pt->val = FX; /* dirichlet */
      else 
	pt->val = VAR; /* neumann */
      return(1);
    }
  return(0);
}

/* caclculate parameter for next point to be inserted to triangulation */
static long double 
param1( long double p1x,long double p1y, long double p2x,long double p2y, long double p3x, long double p3y)
{
  long double temp;
/*assume all are valid point numbers, and none are identical */
  temp =((p1x-p3x)*(p3x-p2x)+(p1y-p3y)*(p3y-p2y))/((p1x-p2x)*(p1y-p3y)-(p1y-p2y)*(p1x-p3x));
  /*  x0 = 0.5*(p1.x+p2.x)+0.5*(p1.y-p2.y)*temp; */
  /* y0 = 0.5*(p1.y+p2.y)-0.5*(p1.x-p2.x)*temp; */
  /* return((y0-p1.y)*(p2.x-p1.x)-(x0-p1.x)*(p2.y-p1.y));*/
#ifdef DEBUG
  printf("param %Lf\n", temp);
#endif
  return(temp);
}

/* a function required to index the edge array */
/* returns the array position of edge p,q */
static int 
idx(int p, int q)
{
  int temp;
  if ( q < p ){/* always q > p */
    temp = p; p = q;q = temp;
  } 
  if ( (p >= 0) && ( p < npoints-1 ) && ( q > 0) && ( q < npoints ) )
    return( (int)((float)p*(float)(npoints)-(float)(p+1)*(float)(p+2)*0.5+q) );
  return(0);

}

#ifdef EXTRA_SOURCE
/* returns other point of p1, except p2 for an edge to exist*/
static int 
other_point( int p1, int p2, position * e_array )
{
  int i, other;
  other = -1; /* invalid point */
  for ( i = 0; i < npoints; i++) {
    if ( (i != p1)&& (i!= p2) )
      if ( e_array[idx(i,p1)] == BOUNDARY )
	other = i;
  }
  return(other);
}
#endif /* ! EXTRA_SOURCE */


#ifdef EXTRA_SOURCE
/* function to find the triangle sharing the edge with points p1 and p2 */
static unsigned long int 
find_correct_neighbour(unsigned long int p1, unsigned long int p2, unsigned long int neigh1, unsigned long int neigh2, unsigned long int neigh3 )
{
  triag * t1;
  t1 =(triag*)findTriangle(neigh1);
  if ( t1 && ( (t1->n1 == p1) || (t1->n2 == p1) || (t1->n3 == p1) ) && ( (t1->n1 == p2) || (t1->n2 == p2) || (t1->n3 == p2) ) )
    return(t1->n);
  t1 =(triag*)findTriangle(neigh2);
  if ( t1 && ( (t1->n1 == p1) || (t1->n2 == p1) || (t1->n3 == p1) ) && ( (t1->n1 == p2) || (t1->n2 == p2) || (t1->n3 == p2) ) )
    return(t1->n);
  t1 =(triag*)findTriangle(neigh3);
  if ( t1 && ( (t1->n1 == p1) || (t1->n2 == p1) || (t1->n3 == p1) ) && ( (t1->n1 == p2) || (t1->n2 == p2) || (t1->n3 == p2) ) )
    return(t1->n);
  return(0);
}
#endif /* ! EXTRA_SOURCE */


/* function to remove intersections of boundary with */
/* triangulation */
int 
remove_crossings(void)
{
  unsigned long int p1,p2,p3,n4,n5,n6,n1,n2,n3;
  triag *t, *neigh1, *neigh2, *neigh3, *neigh4, *neigh5, *neigh6;
  triag *t1, *t2, *t3, told, new1, new2, new3, new4, new5, new6;
  point pt1, pt2, pt3;
  
  unsigned long int i;

  int count;
  int flag = 0;
  
  
  for (  i = 1; i <= triangle_count; i++ ) {
    t =(triag *)findTriangle(i);
#ifdef DEBUG
    printf("considering triangle %ld\n",i);
#endif    
    if ( t ) {
      t1 = t2 = t3 = 0;
      count = 0;
      /* copy original t to told */
      memcpy(&told,t,sizeof(*t));
      /* we consider each edge of the triangle */
      if (  !check_intersection( t->n2, t->n3, edgearray , &pt1 ) ) {
	if ( !check_intersection( t->n3, t->n1, edgearray , &pt1 ) ) {
	  if ( !check_intersection( t->n1, t->n2, edgearray , &pt1 ) ) 
	    count = 0; /* no intersections */
	  else { /* pt1 - (1,2) */
	    count = 1;
	    /* find other detail */
	    n1 = t->n3;
	    n2 = t->n1;
	    n3 = t->n2;
	    t1 = (triag*)findTriangle(t->t3);
	  }
	} else { /* pt1 - (3,1) */
	  /* find details for pt1 */
	  n1 = t->n2;
	  n2 = t->n3;
	  n3 = t->n1;
	  t1 = (triag*)findTriangle(t->t2);
	  if ( ! check_intersection( t->n1, t->n2, edgearray , &pt2 ) ) 
	    count = 1; /* 1 point */
	  else { /* pt1 - (3,1) pt2 - (1,2) */
	    count = 2; /* 2 points */
	    t2 = (triag*)findTriangle(t->t3);
	  }
	}
      } else { /* pt1 - (2,3) */
	if ( !check_intersection( t->n3, t->n1, edgearray , &pt2 ) ) {
	  if ( !check_intersection( t->n1, t->n2, edgearray , &pt2 ) ) {
	    count = 1; /* 1 point */
	    /* find detail */
	    n1 = t->n1;
	    n2 = t->n2;
	    n3 = t->n3;
	    t1 = (triag*)findTriangle(t->t1);
	  } else { /* pt1 - (2,3) pt2 - (1,2) */
	    /* need to interchange the points */
	    memcpy(&pt3,&pt1,sizeof(pt1));
	    memcpy(&pt1,&pt2,sizeof(pt2));
	    memcpy(&pt2,&pt3,sizeof(pt3));
	    /* now pt1 - (1,2) pt2 - (2,3) */
	    n1 = t->n3;
	    n2 = t->n1;
	    n3 = t->n2;
	    t1 = (triag*)findTriangle(t->t3);
	    t2 = (triag*)findTriangle(t->t1);
	    count = 2; /* 2 points */
	  }
	} else { /* pt1 - (2,3) pt2 - (3,1) */
	  /* find detail */
	  n1 = t->n1;
	  n2 = t->n2;
	  n3 = t->n3;
	  t1 = (triag*)findTriangle(t->t1);
	  t2 = (triag*)findTriangle(t->t2);
	  if ( !check_intersection( t->n1, t->n2, edgearray , &pt3 ) ) 
	    count = 2; /* 2 points */
	  else { /*  pt1 - (2,3) pt2 - (3,1) pt3 - (1,2) */
	    count = 3; /* 3 points */
	    t3 = (triag*)findTriangle(t->t3);
	  }
	}
      }
      
      /* now divide according to the number of intersections */
      /* first find out the remaining detail */
      if ( count == 0 ) /* no intersection */
	continue;
      else { /* at least one intersection */
	/* at least pt1 */
	/* first do common things */
	/* split t1 */
	flag = 1;
	pt1.pos = BOUNDARY;
	p1 = insert(pt1);
	if ( t1 ) {
	  if ( t1->n1 == n3 ) {
	    n4 = t1->n3;
	    neigh1 = (triag*)findTriangle(t1->t1);
	    neigh2 = (triag*)findTriangle(t1->t2);
	  } else if ( t1->n2 == n3 ) {
	    n4 = t1->n1;
	    neigh1 = (triag*)findTriangle(t1->t2);
	    neigh2 = (triag*)findTriangle(t1->t3);
	  } else {
	    n4 = t1->n2;
	    neigh1 = (triag*)findTriangle(t1->t3);
	    neigh2 = (triag*)findTriangle(t1->t1);
	  }
	  new1.n1 = n4;
	  new1.n2 = n3;
	  new1.n3 = p1;
	  new1.n = ++triangle_count;
	  new1.t1 = t->n;
	  new1.t2 = t1->n;
	  if ( neigh2 ) {
	    new1.t3 = neigh2->n;
	    if ( neigh2->t1 == t1->n )
	      neigh2->t1 = new1.n;
	    else if ( neigh2->t2 == t1->n )
	      neigh2->t2 = new1.n;
	    else
	      neigh2->t3 = new1.n;
	  } else
	    new1.t3 = 0;
	  insertTriangle(new1);
	  t1->n1 = n4;
	  t1->n2 = p1;
	  t1->n3 = n2;
	  if ( neigh1 ) 
	    t1->t2 = neigh1->n;
	  else
	    t1->t2 = 0;
	  t1->t3 = new1.n;
	}
	
	new3.n = ++triangle_count;
	new3.n1 = p1;
	new3.n3 = n2;
	
	if ( t1 ) {
	   new3.t2 = t1->n;
	   t1->t1 = new3.n;
	   t->t3 = new1.n;
	} else {
	  new3.t2 = 0;
	  t->t3 = 0;
	}
	
	t->n1 = p1;
	t->n2 = n3;
        

	/* all these have to be done */
	/* now consider one intersection */
	if ( count == 1 ) {
	  /* one intersection */
	  if ( n1 == told.n1 ) {
	    neigh3 = (triag*)findTriangle(told.t2);
	    neigh4 = (triag*)findTriangle(told.t3);
	  } else if ( n1 == told.n2 ) {
	    neigh3 = (triag*)findTriangle(told.t3);
	    neigh4 = (triag*)findTriangle(told.t1);
	  } else {
	    neigh3 = (triag*)findTriangle(told.t1);
	    neigh4 = (triag*)findTriangle(told.t2);
	  }
	  
	  t->n3 = n1;
	  new3.n2 = n1;
	  new3.t3 = t->n;
	  t->t2 = new3.n;
	  if ( neigh3 )
	    t->t1 = neigh3->n;
	  else
	    t->t1 = 0;
	  if ( neigh4 ) {
	    new3.t1 = neigh4->n;
	    if ( neigh4->t1 == t->n )
	      neigh4->t1 = new3.n;
	    else  if ( neigh4->t2 == t->n )
	      neigh4->t2 = new3.n;
	    else
	      neigh4->t3 = new3.n;
	  } else
	    new3.t1 = 0;
	  
	  insertTriangle(new3);
	} else { /* 2 or 3 intersections */
	  /* do the common things */
	  /* split t2 */
	  pt2.pos = BOUNDARY;
	  p2 = insert(pt2);
	  if ( t2 ) {
	    if ( t2->n1 == n3 ) {
	      n5 = t2->n2;
	      neigh3 = (triag*)findTriangle(t2->t3);
	      neigh4 = (triag*)findTriangle(t2->t1);
	    } else if ( t2->n2 == n3 ) {
	      n5 = t2->n3;
	      neigh3 = (triag*)findTriangle(t2->t1);
	      neigh4 = (triag*)findTriangle(t2->t2);
	    } else {
	      n5 = t2->n1;
	      neigh3 = (triag*)findTriangle(t2->t2);
	      neigh4 = (triag*)findTriangle(t2->t3);
	    }
	    new2.n1 = p2;
	    new2.n2 = n5;
	    new2.n3 = n1;
	    new2.n = ++triangle_count;
	    if ( neigh4 ) {
	      new2.t1 = neigh4->n;
	      if ( neigh4->t1 == t2->n )
		neigh4->t1 = new2.n;
	      else  if ( neigh4->t2 == t2->n )
		neigh4->t2 = new2.n;
	      else 
		neigh4->t3 = new2.n;
	    } else
	      new2.t1 = 0;
	   
	    new2.t3 = t2->n;
	    t2->n1 = p2;
	    t2->n2 = n3;
	    t2->n3 = n5;
	    if ( neigh3 )
	      t2->t1 = neigh3->n;
	    else
	      t2->t1 = 0;
	    t2->t2 = new2.n;
	    t2->t3 = t->n;
	  }

	  t->n3 = p2;
	  new4.n1 = p2;
	  new4.n2 = n1;
	  new4.n =  ++triangle_count;
	  if ( t2 ) {
	    new4.t3 = new2.n;
	    new2.t2 = new4.n;
	    insertTriangle(new2);
	  } else 
	    new4.t3 = 0;
	  /* all these are common */
	  if ( count == 2 ) {
	    /* 2 intersections */
	    new3.n2 = p2;
	    new4.n3 = n2;
	    new3.t1 = new4.n;
	    new4.t2 = new3.n;
	    if ( n3 == told.n1 )
	      neigh5 =  (triag*)findTriangle(told.t1);
	    else  if ( n3 == told.n2 )
	      neigh5 =  (triag*)findTriangle(told.t2);
	    else
	      neigh5 =  (triag*)findTriangle(told.t3);
	    if ( neigh5 ) {
	      new4.t1 = neigh5->n;
	      if ( neigh5->n1 == n2 )
		neigh5->t3 = new4.n;
	      else if ( neigh5->n2 == n2 )
		neigh5->t1 = new4.n;
	      else 
		neigh5->t2 = new4.n;
	    } else
	       new4.t1 = 0;
	    insertTriangle(new3);
	    insertTriangle(new4);
	  } else if ( count == 3 ) {
	    /* 3 intersections */
	    pt3.pos = BOUNDARY;
	    p3 = insert(pt3);
	    new3.n2 = p3;
	    new4.n3 = p3;
	    new6.n1 = p1;
	    new6.n2 = p2;
	    new6.n3 = p3;
	    new6.n = ++triangle_count;
	    new6.t1 = new4.n;
	    new6.t2 = new3.n;
	    new6.t3 = t->n;
	    new4.t2 = new6.n;
	    new3.t3 = new6.n;
	    t->t2 = new6.n;
	  
	    if (t3) {
	      if ( t3->n1 == n1 ) {
		n6 = t3->n2;
		neigh5 = (triag*)findTriangle(t3->t3);
		neigh6 = (triag*)findTriangle(t3->t1);
	      } else   if ( t3->n2 == n1 ) {
		n6 = t3->n3;
		neigh5 = (triag*)findTriangle(t3->t1);
		neigh6 = (triag*)findTriangle(t3->t2);
	      } else {
		n6 = t3->n1;
		neigh5 = (triag*)findTriangle(t3->t2);
		neigh6 = (triag*)findTriangle(t3->t3);
	      }
	      new5.n1 = p3;
	      new5.n2 = n6;
	      new5.n3 = n2;
	      new5.n = ++triangle_count;
	      if ( neigh6 ) {
		new5.t1 = neigh6->n;
		if ( neigh6->t1 == t3->n )
		  neigh6->t1 = new5.n;
		else if ( neigh6->t2 == t3->n )
		  neigh6->t2 = new5.n;
		else
		  neigh6->t3 = new5.n;
	      } else
		new5.t1 = 0;
	      new5.t2 = new3.n;
	      new5.t3 = t3->n;
	      t3->n1 = p3;
	      t3->n2 = n1;
	      t3->n3 = n6;
	      if ( neigh5 )
		t3->t1 = neigh5->n;
	      else
		t3->t1 = 0;
	      t3->t2 = new5.n;
	      t3->t3 = new4.n;
	      insertTriangle(new5);
	      new3.t1 = new5.n;
	      new4.t1 = t3->n;
	    } else {
	      new3.t1 = 0;
	      new4.t1 = 0;
	    }
	     insertTriangle(new6);
	     insertTriangle(new4);
	     insertTriangle(new3);
	  }
	}
      }
    }
  }
  return(flag);
}


/* function to generate delaunay triangulation */
/* assume : no four points are co-circular */
void 
delaunay( void )
{
 
  edge e, ee; 
  triag tt;
  int i, next;
  long double temp1, temp2;
  position * p_array; /* an array of point position data */
  position * e_array; /* an array of edge position data */
  unsigned int * e_neighbour; /* an array of neighbour triangle numbers of an edge */
  /* to keep track of point being included in the triangulation */
  position lp1, lp2; /* two lines status */
  
  getinput();
  /* now allocate memory to array */
  p_array = ( position *)calloc(npoints, sizeof(position));
  e_array = ( position *)calloc(nedges, sizeof(position));
  e_neighbour =( unsigned int *)calloc(nedges, sizeof(unsigned int));

  /* initialize all points to be out of triangulation */
  for ( i = 0; i < npoints; i++)
    p_array[i] = OUT;
  /* initialize all edges to be not included */
  for( i = 0; i < nedges; i++)
    e_array[i] = OUT;
  /* insert first edge to frontier from hull */
  frontier = E_insert(frontier, hull);
  /* update edges of frontier */
  i = idx(hull.p1, hull.p2);
  e_array[i] = BOUNDARY;
  e_neighbour[i] = 0; /* no triangle */
  p_array[hull.p1] = BOUNDARY;
  p_array[hull.p2] = BOUNDARY;
  /* now loop until there are no edges in the frontier */
  while ( frontier ) {
    //    printf("triangels %d\n",triangle_count);
    frontier = E_get( frontier, &e ); 
    /* check to see whther this has already been inserted to triangulation */
    if (e_array[idx(e.p1,e.p2)] != IN ) {
#ifdef DEBUG     
     printf("line %ld %ld chosen \n", e.p1, e.p2); 
#endif
     /* now analyse this edge */
     /* select the point with lowest parameter circle */
     /* te point should not e already included */
     /* NOTE: this method did not work for some cases */
     /* especially when points are co-circular */
     /* we have to improve this in a future release */      
     next = -1; /* invalid point */
     temp1 = temp2 = -MAX;
     for ( i = 0; i < npoints; i++) {
       if (/*(i!=e.p1)&&(i!=e.p2)&&*/(p_array[i] != IN)&&(ccw1(Mx(e.p1),My(e.p1),Mx(e.p2),My(e.p2),Mx(i),My(i)))/* && (e_array[idx(e.p1,i)] != IN )&&(e_array[idx(e.p2,i)] != IN )*/ ) {
	 //temp2 = param(p[e.p1], p[e.p2],p[i]);/* need to be improved */
	 temp2= param1(Mx(e.p1),My(e.p1),Mx(e.p2),My(e.p2),Mx(i),My(i));
#ifdef DEBUG	 
	 printf("param of %d = %Lf \n", i, temp2);
#endif
	 if ( (temp1 < temp2) /*&& (e_array[idx(e.p1,i)] != IN ) &&(e_array[idx(e.p2,i)] != IN ) */ ) {
	   next = i; /* remember next point */
	   temp1 = temp2;
	 }
       }
     }
     
#ifdef DEBUG
     printf("next %d\n",next);
#endif
     /* now we either have the next point to be added or it does not exist */
     /* if it does not exist , we just remove current edge from frontier */
     /* and go to next one */
     if( next == -1 ) { /* hull edge */ 
       /***************************************************************/
       e_array[idx(e.p1,e.p2)] = IN;
#ifdef DEBUG
       printf("dead end\n");
#endif
       /* shall we do this as well? */
       //p_array[e.p1] = IN;
       //p_array[e.p2] = IN;
     } else {
       /**************************************************************/
       //printf("found next=%d",next);
       lp1 =  e_array[idx(e.p1, next)];
       lp2 =  e_array[idx(next, e.p2)];
	if ( (lp1 == OUT)&&(lp2 == OUT) ) {
	  /*##############################################################*/
#ifdef DEBUG
	    printf("simple case\n");
#endif	  
	    /*____________________________________________________________*/
	    /* simple case, just insert  */
	    /*          *                */
	    /*                           */
	    /*    p1 ______ p2           */
	    /*      /      \             */
	    /* now insert the triangle to mesh */
	    
	    /* the true simple case is the new point is outside*/
	    if ( p_array[next] != OUT ) {
	      /* undo what we did */
	      /* to avoid enclosures within */
	      frontier = E_insert(frontier,e);
#ifdef DEBUG
	      printf("undoing....\n");
#endif	    
	      continue;
	    }
	    tt.n = ++triangle_count;
	    tt.n1 = e.p1;
	    tt.n2 = next;
	    tt.n3 = e.p2;
	    tt.t1 = 0;
	    tt.t2 = e_neighbour[idx(e.p1,e.p2)]; /* neighbour triangle */
	    tt.t3 = 0;
	    insertTriangle( tt );
	    ee.p1 = e.p1;
	    ee.p2 = next;
	    e_neighbour[idx(ee.p1,ee.p2)] = triangle_count;
	    frontier = E_insert( frontier, ee );
	    ee.p1 = next;
	    ee.p2 = e.p2;
	    e_neighbour[idx(ee.p1,ee.p2)] = triangle_count;
	    frontier = E_insert( frontier, ee );
	    /* now update the position of original line */
	    e_array[idx(e.p1, e.p2)] = IN;
	    e_array[idx(e.p1, next)] = BOUNDARY;
	    e_array[idx(next, e.p2)] = BOUNDARY;
	    /* also add new point to frontier */
	    p_array[next] = BOUNDARY;
	    
	  } else if (( lp1 == BOUNDARY ) && ( lp2 == BOUNDARY )) {
	    /* complex case ---*---- */
	    /*                / \    */
	    /*               /   \   */
	    /*           p1 /_____\ p2 */
	    
	    /* three edges will be inserted */
#ifdef DEBUG
	    printf("complex case\n");	    
#endif
	    tt.n = ++triangle_count;
	    tt.n1 = e.p1;
	    tt.n2 = next;
	    tt.n3 = e.p2;
	    tt.t1 = e_neighbour[idx(next,e.p2)]; /* neighbour triangle */
	    tt.t2 = e_neighbour[idx(e.p1,e.p2)];
	    tt.t3 = e_neighbour[idx(e.p1,next)];
	    insertTriangle( tt );
	    e_array[idx(e.p1, e.p2)] = IN;
	    e_array[idx(e.p1, next)] = IN;
	    e_array[idx(next, e.p2)] = IN;
	    p_array[e.p1] = p_array[e.p2] = IN;


	  } else if ( lp1 == BOUNDARY ) {
	    /* lp2 == OUT */
	    /* -------*        */
	    /*       /         */
	    /*    p1/_______ p2 */
	    /*              \   */
#ifdef DEBUG
	    printf("half complex case a\n");
#endif
	    tt.n = ++triangle_count;
	    tt.n1 = e.p1;
	    tt.n2 = next;
	    tt.n3 = e.p2;
	    tt.t1 = 0;
	    tt.t2 = e_neighbour[idx(e.p1,e.p2)];
	    tt.t3 = e_neighbour[idx(next,e.p1)]; /* neighbour triangle */
	    insertTriangle( tt );
	    /* one more edge added to frontier */
	    ee.p1 = next;
	    ee.p2 = e.p2;
	    e_neighbour[idx(ee.p1,ee.p2)] = triangle_count;
	    frontier = E_insert( frontier, ee );
	    e_array[idx(e.p1, e.p2)] = IN;
	    e_array[idx(e.p1, next)] = IN;
	    e_array[idx(next, e.p2)] = BOUNDARY;
	    p_array[e.p1] = IN;
	  } else if ( lp2 == BOUNDARY ) {
	    /* lp1 == OUT */
	    /*           *------------ */
	    /*            \            */
	    /*             \           */
	    /*    p1 _______\ p2       */
	    /*     /                   */
#ifdef DEBUG
	    printf("half complex case b\n");
#endif
	    tt.n = ++triangle_count;
	    tt.n1 = e.p1;
	    tt.n2 = next;
	    tt.n3 = e.p2;
	    tt.t1 = e_neighbour[idx(e.p2,next)]; /* neighbour triangle */
	    tt.t2 = e_neighbour[idx(e.p1,e.p2)];
	    tt.t3 = 0;
	    insertTriangle( tt );
	    /* one more edge added to frontier */
	    /* this is the mirror image of above */
	    ee.p1 = e.p1;
	    ee.p2 = next;
	    e_neighbour[idx(ee.p1,ee.p2)] = triangle_count;
	    frontier = E_insert( frontier, ee );
	    e_array[idx(e.p1, e.p2)] = IN;
	    e_array[idx(e.p2, next)] = IN;
	    e_array[idx(e.p1, next)] = BOUNDARY;
	    p_array[e.p2] = IN;
	  } else {
	    /* unsolvable case */
#ifdef DEBUG
	   
	    printf("next %d p1 %ld p2 %ld\n", next, e.p1,e.p2);
	    printf("line 1 and 2 ");
	    if ( lp1 == IN ) 
	      printf("IN ");
	    else if ( lp1 == OUT )
	      printf("OUT ");
	    else
	      printf("BOUNDARY ");
	     if ( lp2 == IN ) 
	      printf("IN\n");
	    else if ( lp2 == OUT )
	      printf("OUT\n");
	    else
	      printf("BOUNDARY\n");
	     printf("points p1 p2 next ");
	     if ( p_array[e.p1] == IN ) 
	       printf("IN ");
	     else if ( p_array[e.p1] == OUT )
	       printf("OUT ");
	     else
	       printf("BOUNDARY ");
	     if ( p_array[e.p2] == IN ) 
	       printf("IN ");
	     else if ( p_array[e.p2] == OUT )
	       printf("OUT ");
	     else
	       printf("BOUNDARY ");
	     if ( p_array[next] == IN ) 
	       printf("IN\n");
	     else if ( p_array[next] == OUT )
	       printf("OUT\n");
	     else
	       printf("BOUNDARY\n");
#endif
	    
	    /* put the edge back to list */
	     if ( p_array[e.p1] != IN && p_array[e.p2] != IN && lp1 != IN && lp2 != IN ) {
#ifdef DEBUG
	       printf("unsolvable, putting back\n");	       
#endif	    
	       frontier = E_insert( frontier, e );
	     }
	  }
	
      }
    }
  }
  
  
  free(p_array);
  free(e_array);
  free(e_neighbour);
}

