/*  pdnmesh - a 2D finite element solver
    Copyright (c) 2001 Sarod Yatawatta <sarod@cs.pdn.ac.lk>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* mesh with Red Black Tree */
#include <stdio.h>
#include <stdlib.h> /* for malloc(), realloc() */
#include "types.h"

#define NILX &sentinelx /* leaves are sentinel nodes */
node_x sentinelx = { 0, NILX, NILX, NILX, 0, BLACK };
#define NILY &sentinely
node_y sentinely = { 0, 0,NILY, NILY, 0, 0, 0, BLACK, OUT };


Mesh M;
unsigned long int maxpoints;

/* initialise the mesh */
/* __ GLOBAL ___ */
void 
init_mesh(Mesh *M, int maxpoints)
{
  M->Xtree = NILX;
  M->count = 0;
  M->Narray=(node_y**)calloc(maxpoints, sizeof(node_y*));
 
}


/* procedures neede to keep Red-Black tree balanced */
/* first rotation procedures */
/* note we do not use any global variables */
/* __ LOCAL __ */
static void 
rotateLeftx(node_x *x)
{
  /* rotate node x to left */
  node_x *y= x->right;
  /* establish x->right link */
  x->right = y->left;
  if ( y->left != NILX ) y->left->parent = x;

  /* establish y->parent link */
  if ( y != NILX ) y->parent = x->parent;
  if ( x->parent) {
    if ( x == x->parent->left)
      x->parent->left = y;
    else 
      x->parent->right = y;
  } else {
    M.Xtree = y;
  }

  /* link x and y */
  y->left = x;
  if ( x != NILX ) x->parent = y;
}

/* __ LOCAL __ */
static void 
rotateRightx( node_x *x)
{
  /* rotate node x to right */
  node_x *y = x->left;

  /* establish x->left link */
  x->left = y->right;
  if ( y->right != NILX ) y->right->parent = x;

  /* establish y->parent link */
  if ( y != NILX ) y->parent = x->parent;
  if ( x->parent ) {
    if ( x == x->parent->right )
      x->parent->right = y;
    else
      x->parent->left = y;
  } else {
    M.Xtree = y;
  }

  /* link x and y */
  y->right = x;
  if ( x != NILX ) x->parent = y;
}

/* __ LOCAL __ */
static void 
insertFixupx( node_x *x )
{
  /* maintain red-black tree balance after insertion */
  /* check red-black tree properties */
  while ( x != M.Xtree && x->parent->colour == RED ) {
    /* we have a violation */
    if ( x->parent == x->parent->parent->left ) {
      node_x *y = x->parent->parent->right;
      if ( y->colour == RED ) {

	/* uncle is red */
	x->parent->colour = BLACK ;
	y->colour = BLACK;
	x->parent->parent->colour = RED;
	x = x->parent->parent;
      } else {

	/* uncle is black */
	if ( x == x->parent->right ) {
	  /* make x a left child */
	  x = x->parent;
	  rotateLeftx( x );
	}

	/* recolour and rotate */
	x->parent->colour = BLACK;
	x->parent->parent->colour = RED;
	rotateRightx( x->parent->parent );
      }
    } else {
      /* mirror image of above code */
      node_x *y = x->parent->parent->left;
      if ( y->colour == RED ) {

	/* unlce is red */
	x->parent->colour = BLACK;
	y->colour = BLACK;
	x->parent->parent->colour = RED;
	x = x->parent->parent;
      } else {

	/* uncle is black */
	if ( x == x->parent->left ) {
	  x = x->parent;
	  rotateRightx( x );
	}
	x->parent->colour = BLACK;
	x->parent->parent->colour = RED;
	rotateLeftx(x->parent->parent);
      }
    }
  }
  M.Xtree->colour = BLACK;
}

/* now a repetition of the above functions for y node types */
/* __ LOCAL __ */
static void 
rotateLefty(node_y *x, node_x* xparent)
{
  /* rotate node x to left */
  node_y *y= x->right;
  /* establish x->right link */
  x->right = y->left;
  if ( y->left != NILY ) y->left->parent = x;
  
  /* establish y->parent link */
  if ( y != NILY ) y->parent = x->parent;
  if ( x->parent) {
    if ( x == x->parent->left)
      x->parent->left = y;
    else 
      x->parent->right = y;
  } else {
    xparent->Ytree=y;
  }
  
  /* link x and y */
  y->left = x;
  if ( x != NILY ) x->parent = y;
  
}

/* __ LOCAL __ */
static void 
rotateRighty( node_y *x, node_x* xparent )
{
  /* rotate node x to right */
  node_y *y = x->left;
  
  /* establish x->left link */
  x->left = y->right;
  if ( y->right != NILY ) y->right->parent = x;
  
  /* establish y->parent link */
  if ( y != NILY ) y->parent = x->parent;
  if ( x->parent ) {
    if ( x == x->parent->right )
      x->parent->right = y;
    else
      x->parent->left = y;
  } else {
    xparent->Ytree=y;
  }
  
  /* link x and y */
  y->right = x;
  if ( x != NILY ) x->parent = y;
  
}

/* __ LOCAL __ */
static void 
insertFixupy( node_y *x, node_x* xparent )
{
  /* maintain red-black tree balance after insertion */
  /* check red-black tree properties */
  while ( x != xparent->Ytree && x->parent->colour == RED ) {
    /* we have a violation */
    if ( x->parent == x->parent->parent->left ) {
      node_y *y = x->parent->parent->right;
      if ( y->colour == RED ) {
	
	/* uncle is red */
	x->parent->colour = BLACK ;
	y->colour = BLACK;
	x->parent->parent->colour = RED;
	x = x->parent->parent;
      } else {
	
	/* uncle is black */
	if ( x == x->parent->right ) {
	  /* make x a left child */
	  x = x->parent;
	  rotateLefty( x, xparent );
	}

	/* recolour and rotate */
	x->parent->colour = BLACK;
	x->parent->parent->colour = RED;
	rotateRighty( x->parent->parent, xparent );
      }
    } else {
	 /* mirror image of above code */
      node_y *y = x->parent->parent->left;
      if ( y->colour == RED ) {

	/* unlce is red */
	x->parent->colour = BLACK;
	y->colour = BLACK;
	x->parent->parent->colour = RED;
	x = x->parent->parent;
      } else {
	
	/* uncle is black */
	if ( x == x->parent->left ) {
	     x = x->parent;
	     rotateRighty( x, xparent );
	}
	x->parent->colour = BLACK;
	x->parent->parent->colour = RED;
	rotateLefty( x->parent->parent, xparent );
      }
    }
  }
  xparent->Ytree->colour = BLACK;
}


/* the following will insert the given point into the mesh if is already is */
/* not there, else it will do nothing but in both cases, the point number will be returned */
/* ___ GLOBAL ___ */
unsigned long int 
insert( point p )
{/* Mesh M is the global variable */
  node_x *currentx, *parentx, *xx;
  node_y *currenty, *parenty, *yy;
  
  currentx = M.Xtree;
  parentx=0;
     
  /* find the x coordinate first */
  while ( currentx != NILX ) {
    if ( p.x == currentx->x ) {
#ifdef DEBUG
      /* x cordinates match, find y coordinate */
     printf("x match\n"); 
#endif
      currenty = currentx->Ytree;
      parenty=0;
      while ( currenty != NILY ) {
	if ( p.y == currenty->y ) {
#ifdef DEBUG
	  printf("y match as well\n"); 
#endif
	  return( currenty->n);
	}
	parenty = currenty;
	currenty = ( p.y < currenty->y ) ? currenty->left : currenty->right;
      }
      /* if we reach here, we know x corrdinate exist but y does not */
      /* so add new y node to Ytree */
      /* set up new Y node */
      if ( ( yy =( node_y *) malloc( sizeof(node_y) )) == 0 ) {
	printf("not enough memory \n");
	exit(1);
      }
      yy->y = p.y;
      yy->parent = parenty;
      yy->left = NILY;
      yy->right = NILY;
      yy->colour = RED;
      yy->n = M.count; /* increment node count */
      M.Narray[M.count] = yy; /* point it to this */
      yy->xp = &(currentx->x); /* x coordinate */
      yy->pos=p.pos;
      yy->val= p.val;
      /* the potential */
      yy->pot =p.pot;
      /* now insert it in y tree */
      if ( parenty ) {
	if ( p.y < parenty->y )
	  parenty->left = yy;
	else
	  parenty->right = yy;
      } else {
	currentx->Ytree = yy;
	/* certainly we will not reach this step */
	/* because there will be no node with just */
	/* an x value */
      }
#ifdef DEBUG
      printf("new y added\n"); 
#endif
      /* modify global variable */
      insertFixupy( yy, currentx );
 
      /* if root has changed in the function need to do the same here */
      /* if there was no change, just as fine because no change is done */
      if ( M.count >= maxpoints-1 ) {
#ifdef DEBUG	
	printf("point number exceeded, resizing\n");
#endif	
	/* now realloc the array to accomodate bigger size */
	maxpoints +=RESIZE; /* thousand more is enough */
        M.Narray=(node_y**)realloc((void*)M.Narray,maxpoints*sizeof(node_y*));
	if ( M.Narray == 0 ) {
	  printf("not enough memory\n");
	  exit(2);
	}
#ifdef DEBUG
	printf("new size %Ld\n",maxpoints);
#endif      
      }
      /* now increment count */
      M.count++;
      return(yy->n);
      // return(M.count);
    }
    parentx = currentx;
    currentx=( p.x < currentx->x )? currentx->left: currentx->right;
  }
  /* now we know that x coordinate is not in the tree, and */
  /* obviously the y coordinate */
  /* set up new x node */
  if ( (xx =(node_x*) malloc( sizeof(node_x))) == 0 ){
    printf("not enough memory \n");
    exit(1);
  }
  xx->x = p.x;
  xx->parent= parentx;
  xx->left = NILX;
  xx->right= NILX;
  xx->colour = RED;
  /* now insert y node to this x node */
  if ( ( yy =(node_y*) malloc( sizeof(node_y) )) == 0 ) {
    printf("not enough memory \n");
    exit(1);
  }
  yy->y = p.y;
  yy->parent = 0; /* spacial case--should be null-- */
  yy->left = NILY;
  yy->right = NILY;
  yy->colour = RED; /* can make this BLACK and omit calling insertFixupy() */
  yy->n = M.count; /* increment node count */
  M.Narray[M.count] = yy; /* point it to this */
  /* we have a limit to the number of points */
  /* check this before incrementing */
  /* do it at the end */
  yy->pos=p.pos;/* position*/
  yy->pot = p.pot; /* potential */
  yy->val = p.val;
  yy->xp = &(xx->x); /* x coordinate */
  xx->Ytree = yy;
#ifdef DEBUG
  printf("new x and y added\n"); 
#endif
  /* now insert xx in the X tree */
  if ( parentx ) {
    if ( p.x < parentx->x )
      parentx->left = xx;
    else 
      parentx->right = xx;
  } else {
    M.Xtree = xx; /* the first point */
  }
  insertFixupx( xx ); /* we pass globally */
  insertFixupy( xx->Ytree, xx); 
   /* it took me about two weeks to figure out to include the above */
   /* insertFixupy() for the tree to properly work :)               */
   /* this teaches us a lesson: the bugs in a program are NOT where */
   /* you  expect them to be */
   /* note we need not call this function either, we can modify line 364 */
  if ( M.count >= maxpoints-1 ) {  
#ifdef DEBUG
    printf("point number exceeded, resizing\n");
#endif    
    /* now realloc the array to accomodate bigger size */
    maxpoints +=RESIZE; /* thousand more is enough */
    M.Narray=(node_y**)realloc((void*)M.Narray,maxpoints*sizeof(node_y*));
    if ( M.Narray == 0 ) {
      printf("not enough memory\n");
      exit(2);
    }
#ifdef DEBUG
    printf("new size %Ld\n",maxpoints);
#endif 
  }
  M.count++;
  
  return( M.count-1 );
}



/* the following functions are not currently used */
/* however, they may be useful in the future */
/* the following is the converse of insert(), it will return the point */
/* corresponding to the point number  "n" */
/* ___ GLOBAL ___ */

#ifdef EXTRA_SOURCE  
static point 
find_in_mesh( unsigned long int n )
{
  point temp;
#ifdef DEBUG
  printf("n = %ld M.count = %ld\n", n, M.count );
#endif
  /* should this be n >= M.count ??? */
  if ( n > M.count ) {
    printf("no point exists\n");
    return ( temp ); /* does not exist, need not care */
  }
#ifdef DEBUG
  printf("point exists\n"); 
#endif
  temp.y = (M.Narray[n])->y;
  temp.x = *((M.Narray[n])->xp);
  temp.pos = (M.Narray[n])->pos;
  temp.pot =  (M.Narray[n])->pot;
  temp.val =  (M.Narray[n])->val;
  return( temp );
}
#endif /* ! EXTRA_SOURCE */

/*__ LOCAL __ */
/* visit tree in x and y node order */

#ifdef EXTRA_SOURCE  
static void 
visity( node_y *this );/* prototype needed */
static void 
visitx(node_x *this)
{
  if ( this != NILX ) {
    visitx( this->left );
    printf("x = %Lf ", this->x);
    visity( this->Ytree );
    visitx( this->right );
  }
}
#endif /* ! EXTRA_SOURCE */


#ifdef EXTRA_SOURCE  
static void 
visity( node_y *this )
{
  if ( this != NILY ) {
    visity( this->left );
    printf("no= %ld y = %Lf\n", this->n,this->y);
    //printf(" y = %Lf\n", this->y);  
    visity( this->right );
  }
}
#endif /* ! EXTRA_SOURCE */

/* ___ GLOBAL ___ */

#ifdef EXTRA_SOURCE  
static void 
visit(Mesh M)
{
  node_x *currentx;
  currentx = M.Xtree;
  if ( currentx->left != NILX) visitx(currentx->left);
  printf("x = %Lf ", currentx->x); 
  visity(currentx->Ytree);
  if ( currentx->right != NILX) visitx(currentx->right);
}
#endif /* ! EXTRA_SOURCE */

/* __ LOCAL __ */

#ifdef EXTRA_SOURCE  
static void 
destroyXtree( node_x* xx )
{
   if ( xx->right != NILX )
     destroyXtree( xx->right );
   if ( xx->left != NILX )
     destroyXtree( xx->left );
   free( xx);
}
#endif /* ! EXTRA_SOURCE */

/* ___ GLOBAL */

#ifdef EXTRA_SOURCE  
static void 
destroy_mesh( Mesh M )
{ /* free up memory */
   int i;
   for ( i =0; i < M.count; i++)
     free( M.Narray[i] ); /* free up the y nodes */
  
    free( M.Narray );
   destroyXtree( M.Xtree );
/*   free( M.Xtree); */
}
#endif /* ! EXTRA_SOURCE */
