/*  pdnmesh - a 2D finite element solver
    Copyright (c) 2001 Sarod Yatawatta <sarod@cs.pdn.ac.lk>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <stdio.h>
#include <math.h> /* for sqrt() */
#include "types.h"

#ifndef sqrtf
#define sqrtf(x) sqrt(x)
#endif


unsigned long int * renumber; /* re-numbering array */
unsigned long int * re_renumber; /* re-renumbering array */
/* used to map potentials back to the original points */

/* maximum and minimum poteltials for contours */
long double maxpot;
long double minpot;

long double splitter; /* to set triangle priority */



/* a function to solve the matrix equation by conjugate gradient */
/* method */
static void 
solve1( long double **a, long double *x, long double *y, int N) 
{ /* solves a system of equations [a][x] = [b] using conjugate gradient
     method
  */
  long double *p, /* parameter */
    *r, /*residuals */
    *temp; /* temp array */
  
  long double prod1, /* temp variables */
    alpha,
    beta,
    e,
    olde,
    toelarance;
  
  unsigned long int i,j,iter=0,max_iter =1000;
   toelarance = 0.00001; /* squared */
   
  
  /* allocate memory */
  if ( (p = malloc((size_t)N*sizeof(long double))) == 0 ) {
    printf("solve: outof memory");
    exit(1);
  }
  if ( (r = malloc((size_t)N*sizeof(long double))) == 0 ) {
    printf("solve: outof memory");
    exit(1);
  }						 
  if ( (temp = malloc((size_t)N*sizeof(long double))) == 0 ) {
    printf("solve: outof memory");
    exit(1);
  }
#ifdef DEBUG 
  printf("********************\n");
  for ( i = 0; i < N; i++){
    printf("|  ");
    for( j = 0; j < N; j++)
      printf(" %Lf  ", a[i][j]);
    printf("=  %Lf  |\n", y[i]);
  } 
#endif
  /* preconditioning */
  /* first make the diagonal elements square root */
  for ( i = 0; i < N; i++ )
    if ( a[i][i] > 0 )
      a[i][i] = sqrtf(a[i][i]);
    else {
      fprintf(stderr,"warning: bad matrix\'");
    }
  
  
  /* now divide all other elements */
  /* since the matrix is symmetric, divide only half */
  for ( i =0; i < N; i++ )
    for ( j = 0; j < i; j++ ) {
      a[i][j] = a[i][j]/(a[i][i]*a[j][j]);
      a[j][i] = a[i][j];
    }
  
  /* now modify the potentials */
  for ( i = 0; i < N; i++ )
    y[i] /= a[i][i]; 
  
  
  /* initial guess */
  for ( i = 0; i <  N; i ++ )
    x[i] = 0;
  
  /* initial residual */
  for ( i = 0; i < N; i++ ) {
    
    prod1 = 0;
    /* avoid using a[i][i] in product */
    for ( j = 0; j < i ; j++ )
      prod1 += a[i][j]*x[j];
    prod1 += x[i]; /* skip a[i][i] */
    for ( j = i+1; j < N ; j++ )
      prod1 += a[i][j]*x[j];
    
    r[i] = y[i] - prod1;
#ifdef DEBUG
    printf("r[%ld] = %Lf\n",i, r[i]);
#endif
    p[i] = r[i];
  }
  
  e = 0; /* reset error */
  for ( i = 0; i < N; i++ ) 
    e += r[i]*r[i]; /* norm squared */
  
#ifdef DEBUG
  printf("error = %Lf ", e);
#endif
  olde = e; 
  
 
  
  while ( (olde > toelarance*e) && (iter < max_iter) ) {
    prod1 = 0;
    for ( i =0; i < N; i++ ) {
      temp[i] = 0;
      /* avoid using a[i][i] */
      for ( j = 0; j < i; j++ )
	temp[i] += a[i][j]*p[j];
      temp[i] += p[i]; 
      for ( j = i+1; j < N; j++ )
	temp[i] += a[i][j]*p[j];
      
      prod1 += p[i]*temp[i];
    }
    
    alpha = e/prod1;
#ifdef DEBUG
    printf("alpha = %Lf ",alpha);
#endif
    
    
    for ( i =0; i < N; i++ ) 
      x[i] += alpha*p[i];
    
    if ( !iter%50 ) {  
      for ( i = 0; i < N; i++ ) {
	prod1 = 0;
	/* avoid using a[i][i] in product */
	for ( j = 0; j < i ; j++ )
	  prod1 += a[i][j]*x[j];
	prod1 += x[i]; /* skip a[i][i] */
	for ( j = i+1; j < N ; j++ )
	  prod1 += a[i][j]*x[j];     
	
	r[i] = y[i] - prod1;
      }
    } else {
      for ( i = 0; i < N; i++ )
	r[i] -= alpha*temp[i];
      
    }   
    
    olde = e;
    e = 0;
    for ( i = 0; i < N; i++ )
      e += r[i]*r[i];
#ifdef DEBUG
    printf("error = %Lf\n",e);
#endif   
    beta = e/olde;
    
    
    for ( i = 0; i < N; i++ ) 
      p[i] = r[i] - beta*p[i];
    
#ifdef DEBUG
    for ( i =0; i < N; i++ ) {
      printf("x[%ld] = %Lf\n",i,x[i]);
    }
#endif
    
    iter++; 
  } 
  
  free(p);
  free(r);
  free(temp);
  /* now modify x[] */
  for ( i =0; i < N; i++ )
    x[i] /= a[i][i]; 
  
}


/* a function to solve a matrix equation by symmetric LU */
/* decomposition or cholevsky decomposition */
static void 
solve( long double **a, long double *x, long double *y, int N) 
{
  /* solves a system in the form [a][x] = [y] order N  where [a] is symmetric*/
  /* first cholevsky decomposition */
  int i,j,k;
  long double temp;

#ifdef DEBUG
  for ( i = 0; i < N ; i++ ) {
    for ( j = 0 ; j <= i; j++ )
      printf("%Lf ",a[i][j]);
    printf("\n");
  }
#endif

  for ( j = 0; j < N; j++) {
    /* first diagonal element */
    temp = 0;
    for ( k = 0; k < j; k++) 
      temp += a[j][k]*a[j][k];
    a[j][j] = sqrtf( a[j][j] - temp );
#ifdef DEBUG
    printf("col %d diag %Lf\n",j,a[j][j]);
#endif
    /* check for singularities */
    if ( a[j][j] <= 0.0f ) {
      a[j][j] = TOL; /* a small value */
      fprintf(stderr,"warning: solve: singular matrix\n");
    }
    /* now handle the i th column */
    for( i = j+1; i < N; i++) {
	  temp = 0;
	for ( k = 0; k < j; k++)
	  temp += a[i][k]*a[j][k];
	a[i][j]  = ( a[i][j] - temp )/a[j][j];
#ifdef DEBUG
	printf("row %d elem %Lf\n",i,a[i][j]);
#endif
    }
  }
  
#ifdef DEBUG
  for ( i =0; i < N ; i++ ) {
    for ( j = 0 ; j <= i; j++ )
      printf("%Lf ",a[i][j]);
    printf("\n");
  }
#endif
 
  /* forward elimination */
  for ( i = 0; i < N; i++) {
    temp = 0;
    for ( k = 0; k < i; k++)
      temp += a[i][k]*y[k];
	 /* since we have eliminated singularities, no need to check a[i][i]*/
      y[i] = ( y[i] - temp )/a[i][i];
  }

  /* backward substitution */
  for ( i = N-1; i >= 0; i--) {
    temp = 0;
    for ( j = i+1; j< N; j++)
      temp += x[j]*a[j][i];
	 /* since we have eliminated singularities, no need to check a[i][i]*/
      x[i] = ( y[i] - temp )/a[i][i];
  }
}


/* a function to build the global matrix or the matrix */
/* equation for the system */
static void 
build( long double **P, long double *q, long double *x, long double *y, long double *phi, int NU, int NE, int NN) 
{
  /* infd - input file descriptor                  */
  /* P[][] - global matrix to build   NUxNU        */
  /* q[] - global array to build     NUx1          */
  /* x[] - x coordinates             NNx1          */
  /* y[] - y coordinates             NNx1          */
  /* phi[] - potentials              NNx1          */
  /* NU - no of unknown potential nodes             */
  /* NE - no of elements                            */
  /* NN - no of nodes                               */

  /* now the running variables for each triangular element */
  int v[3]; /* to keep vertex number of each node */
  long double pl[3][3];  /* local version of P[][] above */
  long double ql[3];	/* local version of q[] above */
  long double rho[3]; /* rho values */ /* conductivity */
  triag *t;
 
  int    i,j,id;
  
  long double 
    epsilon, /* permittivity */
    A, /* area of a triangulr element */
    k, /* value of the determinanat */
    yy0, yy1, yy2;
  
  long double b[3]; /*local arrays */
  long double c[3];

  /* just to be on the safe side */
  

	/* now read node data with unknown potentials */
  for ( i = 0; i < NU; i++) {
    x[i]= (Mx(re_renumber[i]))*scale-xoffset;
    y[i]= (My(re_renumber[i]))*scale-yoffset;
  }
	
  /* now read node data with known potentials , which comes last */
  for ( i = NU; i < NN; i++ ) { 
    x[i] = (Mx(re_renumber[i]))*scale-xoffset;
    y[i] = (My(re_renumber[i]))*scale-yoffset;
    phi[i] = (Mpot(re_renumber[i]));
  }
	
  /* now initialize P[][] and q[] */
  for ( i = 0; i < NU; i++) {
    q[i] = 0;
    for ( j = 0; j <= i; j++) 
      P[i][j] = 0;
  }
  
#ifdef DEBUG
	for  ( i = 0; i< NN; i++)
	    printf("%d %Lf, %Lf\n", i, x[i], y[i]); 
#endif
	
	/* now read element data, whilst building the global matrices */
	for ( id = 0; id < triangle_count; id++) {
	  t =(triag *)findTriangle(id);
	  if ( t ) {
	    
	    v[0] = renumber[t->n3];
	    v[1] = renumber[t->n2];
	    v[2] = renumber[t->n1]; /* ccw direction */
	    rho[0] = ex( boundarray[t->boundary].rhoexp, t->n1);
	    rho[1] = ex( boundarray[t->boundary].rhoexp, t->n2);
	    rho[2] = ex( boundarray[t->boundary].rhoexp, t->n3);
	   
	    epsilon = t->epsilon;
	    
	    //	    printf("epsilon %Lf rho %Lf\n", epsilon, rho);

	    /* now have read one triangle , add it to global */
	    /* first the coefficients */
	    /* some temporary values, neede several times */
	    yy0 = y[v[1]]-y[v[2]];
	    yy1 = y[v[2]]-y[v[0]];
	    yy2 = y[v[0]]-y[v[1]];
	    
	    /* area of the triangle */
	    A = 0.5*( x[v[0]]*yy0 + x[v[1]]*yy1 + x[v[2]]*yy2 ); 
	    A = fabsl(A);
	    /* another constant */
	    k = 0.5/A;
	    //k = 1;
	    /* now calculate b's and c's */
#ifdef DEBUG
	  printf("adding element with nodes %d %d and %d...\n\n", v[0], v[1], v[2]);
#endif
	  b[0]= k*yy0;
#ifdef DEBUG
	  printf("b0=%Lf  ", b[0]);
#endif
	  b[1]= k*yy1;
#ifdef DEBUG
	  printf("b1=%Lf  ", b[1]);
#endif
	  b[2]= k*yy2;
#ifdef DEBUG
	  printf("b2=%Lf \n", b[2]);
#endif	  
	  c[0] = k*( x[v[2]]-x[v[1]] ); 
#ifdef DEBUG
	  printf("c0=%Lf  ", c[0]); 
#endif
	  c[1] = k*( x[v[0]]-x[v[2]] );
#ifdef DEBUG
	  printf("c1=%Lf  ", c[1]);
#endif
	  c[2] = k*( x[v[1]]-x[v[0]] );
#ifdef DEBUG
	  printf("c2=%Lf \n", c[2]);
#endif
	  
	  /* now build pl[] and ql */
	  for ( i = 0; i < 3; i++ ) 	
	    for ( j = 0; j < 3; j++ )
	      pl[i][j] = A*epsilon*(b[i]*b[j]+c[i]*c[j]); 
	  /* I thought it should be epsilon/3 here	*/
	  
	  //ql = epsilon* A /3 ;
	  ql[0] = A*(2*rho[0]+rho[1]+rho[2])/12.0;
	  ql[1] = A*(rho[0]+2*rho[1]+rho[2])/12.0;
	  ql[2] = A*(rho[0]+rho[1]+2*rho[2])/12.0;
	  /* Here instead should not rho go ? */
#ifdef DEBUG 
	  printf("local p and q ..\n");
	  for ( i = 0; i <3; i++) {
	    printf("| ");
	    for ( j = 0; j < 3; j++)
	      printf(" %Lf ", pl[i][j]);
	    printf("|| %Lf  | \n", ql[i]); 
	  }	
#endif 
	  
	  
	  /* better to use multiplication than division */
	  /* now augment the global data set */
	  for ( i = 0; i < 3; i++)
	    if ( v[i] < NU ) { /* we index from 0, not from 1 */
	      q[v[i]] = q[v[i]] + ql[i];
	      for ( j = 0; j < 3; j++ ) {
		if ( v[j] < NU )
		  { 
					 if ( v[j] <= v[i] )
					 P[v[i]][v[j]] = P[v[i]][v[j]] + pl[i][j];
					 /* we only build half since we have half only */
		  }
		else 
		  { 
					 q[v[i]] = q[v[i]] - (pl[i][j])*phi[v[j]];
		  } 
	      }
	    } /*else printf("rejecting %d\n", v[i]);*/
#ifdef DEBUG
	  printf("building global matrix ..\n");
#endif
	  }
	}
	
#ifdef DEBUG	  
	  for ( i = 0; i < NU; i++){
	    printf("|  ");
	    for( j = 0; j <= i; j++)
	      printf(" %Lf  ", P[i][j]);
	    printf("=  %Lf  |\n", q[i]);
	    } 
	   for ( i = 0; i < NN; i++) 
		printf("node %d potential %Lf\n", i, phi[i]); 
#endif
}


/* after genarating the mesh, need to re-number points */
/* so that unknown points are first */
/* then we solve the problem */
void 
re_number_and_solve(void)
{
  unsigned long int i,j;
 
  unsigned long int unknowns=0; /* unknown points */
  unsigned long int total = 0; /* total points */
  unsigned long int triangles = 0; /* known points */


  triag *t;
  
  /* first the global node data, all total size */
  long double *x ; /* x coordinates */
  long double *y ; /* y coordinates */
  long double *phi ; /* potentials at each node */
  
  /* the global variables */
  long double **P ; /* size unknownsxunknowns note we do not need a NN x NN matrix, only the unknowns
		     */
  long double *q ; /* unknowns sizesame here */

  /* allocate memory for arrays */
  renumber = (unsigned long int *)calloc(M.count,sizeof(unsigned long));
  re_renumber = (unsigned long int *)calloc(M.count,sizeof(unsigned long));

  /* need to solve for only points referred by triangles */
  /* SO -- */
  for ( i =1; i <= triangle_count; i++ ) {
    t =(triag *)findTriangle(i);
    if ( t ) {
      /* mark the points in this triangle */
      /* use re_renumber[] array for this */
      re_renumber[t->n1] = 1;
      re_renumber[t->n2] = 1;
      re_renumber[t->n3] = 1;
    }
  }

  /* now find unknowns, or INSIDE points */
  j = 0;
  for ( i = 0; i < M.count; i++ )
    if (  Mval(i) == VAR && re_renumber[t->n1] == 1 ) {
      renumber[i] = j++;
      unknowns++;
    }
  /* now renumber the known points */
  total = unknowns;
  for ( i = 0; i < M.count; i++ )
    if ( Mval(i) == FX && re_renumber[t->n1] == 1  ) {
      renumber[i] = j++;
      total++;
    }
  
  /* now the number of actual triangles */
  for ( i =1; i <= triangle_count; i++ ) {
    t =(triag *)findTriangle(i);
    if ( t )
      triangles++;
  }
  

  
  /* now construct re-renmbering array */
  i = 0;
  /* do an insertion sort */
  while ( i < M.count ) {
    for ( j = 0; j < M.count; j++ ) 
      if ( renumber[j] == i )
	break;
    re_renumber[i] = j;
    i++;
  }
  
#ifdef DEBUG
  printf("i,  renum, rerenum\n");
  for ( i = 0; i < M.count; i++ )
    printf("%ld %ld %ld\n",i,renumber[i],re_renumber[i]);
  for ( i = 0; i < total; i++ )
    printf("%ld  %Lf  %Lf %Lf\n",i,Mx(i),My(i),Mpot(i));
#endif
  
  /* we free these two arrays only after solving the problem */
  /* now solve the problem */
  /* first allocate memory for the arrays */
  x=(long double *)calloc(total, sizeof(long double));
  y=(long double *)calloc(total, sizeof(long double));
  phi=(long double *)calloc(total, sizeof(long double));
  q=(long double *)calloc(unknowns, sizeof(long double));
   
	
	/* now the matrix */
	/* STEP 1: SET UP THE ROWS. */
	P= (long double **) calloc(unknowns , sizeof(long double *));
	
	/* STEP 2: SET UP THE COLUMNS. */
	for (i = 0; i < unknowns; ++i)
			  /* variable length columns */
	  P[i] = ( long double * ) calloc( (size_t)i+1, sizeof(long double));
	
	/* now the memory allocation is over */
  (void)build( P, q, x, y, phi, unknowns, triangles, total);
  (void)solve(P,phi, q, unknowns);

  /* after solution, array phi[] will contain the new potentials */
  /* now put them back to the point data */
  /* we also note the maximum and minimum values for contour plotting */
  maxpot = -1000;
  minpot = 1000;
  for ( i = 0; i < total; i++ ) {
    Mpot(re_renumber[i]) = phi[i];
    if ( phi[i] > maxpot )
      maxpot = phi[i];
    if ( phi[i] < minpot )
      minpot = phi[i];
  }



  /* now we can free the renumber[] array */
  free(renumber);
  /* now free this array too */
  free(re_renumber);
  
  /* free everything else */
  
  /* first the arrays */
  free(x);
  free(y);
  free(phi);
  free(q);
  
  /* next the matrix */
  for ( i = 0; i < unknowns ; i++)
    free(P[i]);
  free(P);

 
  /* now printout a summery */
#ifdef DEBUG
  for ( i = 0; i < total; i++ )
    printf("%ld  %Lf  %Lf %Lf\n",i,Mx(i),My(i),Mpot(i));
#endif

}

/* function to generate a file containing the mesh */
void 
write_mesh_file(char * file_name)
{

  unsigned long int i,j;
 
  unsigned long int unknowns=0; /* unknown points */
  unsigned long int total = 0; /* total points */
  unsigned long int triangles = 0; /* known points */

  FILE * outfd; /* output file */

  long double rho; /* calculated rho */
  triag *t;

  /* allocate memory for arrays */
  renumber = (unsigned long int *)calloc(M.count,sizeof(unsigned long));
  re_renumber = (unsigned long int *)calloc(M.count,sizeof(unsigned long));

  /* need to solve for only points referred by triangles */
  /* SO -- */
  for ( i =1; i <= triangle_count; i++ ) {
    t =(triag *)findTriangle(i);
    if ( t ) {
      /* mark the points in this triangle */
      /* use re_renumber[] array for this */
      re_renumber[t->n1] = 1;
      re_renumber[t->n2] = 1;
      re_renumber[t->n3] = 1;
    }
  }

  /* now find unknowns, or INSIDE points */
  j = 0;
  for ( i = 0; i < M.count; i++ )
    if (  Mval(i) == VAR && re_renumber[t->n1] == 1 ) {
      renumber[i] = j++;
      unknowns++;
    }
  /* now renumber the known points */
  total = unknowns;
  for ( i = 0; i < M.count; i++ )
    if ( Mval(i) == FX && re_renumber[t->n1] == 1  ) {
      renumber[i] = j++;
      total++;
    }
  
  /* now the number of actual triangles */
  for ( i =1; i <= triangle_count; i++ ) {
    t =(triag *)findTriangle(i);
    if ( t )
      triangles++;
  }
  

  
  /* now construct re-renmbering array */
  i = 0;
  /* do an insertion sort */
  while ( i < M.count ) {
    for ( j = 0; j < M.count; j++ ) 
      if ( renumber[j] == i )
	break;
    re_renumber[i] = j;
    i++;
  }
  
     /* problem? what if some points are not included in any triangle ? */
#ifdef DEBUG
  printf("total %ld unknowns %ld knowns %ld triangles %ld\n", M.count, unknowns,total-unknowns,triangles);
#endif
  
  /* open output file */
  outfd =fopen( file_name , "w" );
  if ( outfd == NULL ) {
    fprintf(stderr, "could not open file %s\n", file_name);
    exit(1);
  }

  fprintf(outfd,"#\t automatically generated by "PACKAGE" "VERSION"\n");
  /* now print the output in standard form */
  /* first row, total numberof nodes, triangles, unknown points */
  fprintf(outfd, "%ld %ld %ld\n",total, triangles, unknowns);
  /* now print the unknown point coords, with the number */
  for ( i = 0; i < unknowns; i++ )
    fprintf(outfd,"%ld %Lf %Lf\n",i, (Mx(re_renumber[i]))*scale-xoffset, (My(re_renumber[i]))*scale-yoffset );
  /* now the known coordinates with potentials */
  for ( i = unknowns; i < total; i++ )
    fprintf(outfd,"%ld %Lf %Lf %Lf\n",i,  (Mx(re_renumber[i]))*scale-xoffset , (My(re_renumber[i]))*scale-yoffset, Mpot(re_renumber[i]));
  /* now the triangles data with rho and epsilon */
  for ( i =1; i <= triangle_count; i++ ) {
    t =(triag *)findTriangle(i);
    if ( t ) {
	    rho = (ex( boundarray[t->boundary].rhoexp, t->n1) 
					+ex(boundarray[t->boundary].rhoexp, t->n2)
					+ex(boundarray[t->boundary].rhoexp, t->n3) )*0.333333333;
      fprintf(outfd,"%ld %ld %ld %ld %Lf %Lf\n",i, renumber[t->n3],renumber[t->n2],renumber[t->n1]
	      ,t->epsilon, rho);
	 }
  }
  
  /* now close file */
  fclose(outfd);

  
}

/* a function update priorities of each 
   triangle using the potential values */
void 
update_priorities(void)
{
  unsigned long int i;
  triag *t;

  for( i =0; i < triangle_count; i++ ) {
    t =(triag *)findTriangle(i);
    if ( t ) {
      /* calculate gradient of three edges */
      t->par =(fabsl(Mpot(t->n1)-Mpot(t->n2))
	+fabsl(Mpot(t->n2)-Mpot(t->n3))
	+fabsl(Mpot(t->n3)-Mpot(t->n1)) )*splitter;
#ifdef DEBUG
      printf("priority of %ld=%Lf\n",i,t->par);
#endif
    }
  }
}

