/*  pdnmesh - a 2D finite element solver
    Copyright (c) 2001 Sarod Yatawatta <sarod@cs.pdn.ac.lk>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif



/* red-black tree of triangles */
#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* for memcpy() */
#include <stdarg.h>
#include "types.h"
#include <GL/glut.h> /* for drawing function */

#define NIL &sentinel           /* all leafs are sentinels */
Node sentinel = { NIL, NIL, 0, BLACK, 0};



Node *root = NIL;               /* root of Red-Black tree */



static void 
rotateLeft(Node *x) {

   /**************************
    *  rotate node x to left *
    **************************/

    Node *y = x->right;

    /* establish x->right link */
    x->right = y->left;
    if (y->left != NIL) y->left->parent = x;

    /* establish y->parent link */
    if (y != NIL) y->parent = x->parent;
    if (x->parent) {
        if (x == x->parent->left)
            x->parent->left = y;
        else
            x->parent->right = y;
    } else {
        root = y;
    }

    /* link x and y */
    y->left = x;
    if (x != NIL) x->parent = y;
}

static void 
rotateRight(Node *x) {

   /****************************
    *  rotate node x to right  *
    ****************************/

    Node *y = x->left;

    /* establish x->left link */
    x->left = y->right;
    if (y->right != NIL) y->right->parent = x;

    /* establish y->parent link */
    if (y != NIL) y->parent = x->parent;
    if (x->parent) {
        if (x == x->parent->right)
            x->parent->right = y;
        else
            x->parent->left = y;
    } else {
        root = y;
    }

    /* link x and y */
    y->right = x;
    if (x != NIL) x->parent = y;
}

static void 
insertFixup(Node *x) {

   /*************************************
    *  maintain Red-Black tree balance  *
    *  after inserting node x           *
    *************************************/

    /* check Red-Black properties */
    while (x != root && x->parent->color == RED) {
        /* we have a violation */
        if (x->parent == x->parent->parent->left) {
            Node *y = x->parent->parent->right;
            if (y->color == RED) {

                /* uncle is RED */
                x->parent->color = BLACK;
                y->color = BLACK;
                x->parent->parent->color = RED;
                x = x->parent->parent;
            } else {

                /* uncle is BLACK */
                if (x == x->parent->right) {
                    /* make x a left child */
                    x = x->parent;
                    rotateLeft(x);
                }

                /* recolor and rotate */
                x->parent->color = BLACK;
                x->parent->parent->color = RED;
                rotateRight(x->parent->parent);
            }
        } else {

            /* mirror image of above code */
            Node *y = x->parent->parent->left;
            if (y->color == RED) {

                /* uncle is RED */
                x->parent->color = BLACK;
                y->color = BLACK;
                x->parent->parent->color = RED;
                x = x->parent->parent;
            } else {

                /* uncle is BLACK */
                if (x == x->parent->left) {
                    x = x->parent;
                    rotateRight(x);
                }
                x->parent->color = BLACK;
                x->parent->parent->color = RED;
                rotateLeft(x->parent->parent);
            }
        }
    }
    root->color = BLACK;
}

static Node *
insertNode(triag t) {
    Node *current, *parent, *x;

   /***********************************************
    *  allocate node for data and insert in tree  *
    ***********************************************/

    /* find where node belongs */
    current = root;
    parent = 0;
    while (current != NIL) {
        if (t.n == current->T->n) return (current);
        parent = current;
        current = (t.n < current->T->n) ?
            current->left : current->right;
    }

    /* setup new node */
    if ((x = malloc (sizeof(*x))) == 0) {
        printf ("insufficient memory (insertNode)\n");
        exit(1);
    }
    if ((x->T = malloc(sizeof(*x->T)))== 0) {
      printf ("insufficient memory (insertNode)\n");
      exit(1);
    }
    
    memcpy((void*)x->T, &t, sizeof(t));
    
    x->parent = parent;
    x->left = NIL;
    x->right = NIL;
    x->color = RED;

    /* insert node in tree */
    if(parent) {
        if(t.n < parent->T->n)
            parent->left = x;
        else
            parent->right = x;
    } else {
        root = x;
    }

    insertFixup(x);
    return(x);
}


static int 
updateNeighbour( Node * n, unsigned long t_number, unsigned long pt1, unsigned long pt2 )
{
  /* n is not NULL */
  /* find common points to insert in correct place */
  /* pt1, pt2 are the numbers common to both triangles */
  if ( ((n->T->n2 == pt1) && (n->T->n3 == pt2))||((n->T->n2 == pt2) && (n->T->n3 == pt1)) ) {
    /* n1 is the non common point */
    n->T->t1 = t_number;
    return(0);
  }
  if ( ((n->T->n1 == pt1) && (n->T->n3 == pt2))||((n->T->n1 == pt2) && (n->T->n3 == pt1))  ) {
    /* n2 is the non common point */
    n->T->t2 = t_number;
    return(0);
  }
  if ( ((n->T->n1 == pt1) && (n->T->n2 == pt2))||((n->T->n1 == pt2) && (n->T->n2 == pt1)) ) {
    /* n3 is the non common point */
    n->T->t3 = t_number;
    return(0);
  }
  return(-1); /* error */
}


static void 
deleteFixup(Node *x) {

   /*************************************
    *  maintain Red-Black tree balance  *
    *  after deleting node x            *
    *************************************/

    while (x != root && x->color == BLACK) {
        if (x == x->parent->left) {
            Node *w = x->parent->right;
            if (w->color == RED) {
                w->color = BLACK;
                x->parent->color = RED;
                rotateLeft (x->parent);
                w = x->parent->right;
            }
            if (w->left->color == BLACK && w->right->color == BLACK) {
                w->color = RED;
                x = x->parent;
            } else {
                if (w->right->color == BLACK) {
                    w->left->color = BLACK;
                    w->color = RED;
                    rotateRight (w);
                    w = x->parent->right;
                }
                w->color = x->parent->color;
                x->parent->color = BLACK;
                w->right->color = BLACK;
                rotateLeft (x->parent);
                x = root;
            }
        } else {
            Node *w = x->parent->left;
            if (w->color == RED) {
                w->color = BLACK;
                x->parent->color = RED;
                rotateRight (x->parent);
                w = x->parent->left;
            }
            if (w->right->color == BLACK && w->left->color == BLACK) {
                w->color = RED;
                x = x->parent;
            } else {
                if (w->left->color == BLACK) {
                    w->right->color = BLACK;
                    w->color = RED;
                    rotateLeft (w);
                    w = x->parent->left;
                }
                w->color = x->parent->color;
                x->parent->color = BLACK;
                w->left->color = BLACK;
                rotateRight (x->parent);
                x = root;
            }
        }
    }
    x->color = BLACK;
}

static void 
deleteNode(Node *z) {
    Node *x, *y;

   /*****************************
    *  delete node z from tree  *
    *****************************/

    if (!z || z == NIL) return;


    if (z->left == NIL || z->right == NIL) {
        /* y has a NIL node as a child */
        y = z;
    } else {
        /* find tree successor with a NIL node as a child */
        y = z->right;
        while (y->left != NIL) y = y->left;
    }

    /* x is y's only child */
    if (y->left != NIL)
        x = y->left;
    else
        x = y->right;

    /* remove y from the parent chain */
    x->parent = y->parent;
    if (y->parent)
        if (y == y->parent->left)
            y->parent->left = x;
        else
            y->parent->right = x;
    else
        root = x;

    /* here the original triangle is copied ? */
    if (y != z) z->T = y->T;


    if (y->color == BLACK)
        deleteFixup (x);
    /* do not free the triangle, it now belongs to another node */
    if ( y == z )  free(y->T);
    /* else do not free */
    free(y);
}

static Node *
findNode(unsigned long t_number) {

   /*******************************
    *  find node containing data  *
    *******************************/

    Node *current = root;
    while(current != NIL)
        if(t_number == current->T->n)
            return (current);
        else
            current = (t_number <  current->T->n) ?
                current->left : current->right;
    return(0);
}

/* inserts a triangle to tree */
void 
insertTriangle( triag t )
{
  Node * temp;
  /* first find the neighbours of the triangle */
  /* all triangles start from number 1 */
  if ( t.t1 > 0 ) {
    temp = findNode( t.t1 );
    /* if temp not NULL or SENTINEL , neighbour exists */
    if ( temp != 0 )
      updateNeighbour( temp , t.n , t.n2, t.n3);
  }
  if ( t.t2 > 0 ) {
    temp = findNode( t.t2 );
    /* if temp not NULL or SENTINEL , neighbour exists */
    if ( temp != 0 )
      updateNeighbour( temp , t.n , t.n1, t.n3);
  }
  if ( t.t3 > 0 ) {
    temp = findNode( t.t3 );
    /* if temp not NULL or SENTINEL , neighbour exists */
    if ( temp != 0 )
      updateNeighbour( temp , t.n , t.n1, t.n2);
  }

  /* now neighbours are updated, insert the triangle */
  insertNode(t);
}

static int 
backdateNeighbour( Node * n, unsigned long t_number )
{
  /* n is not NULL */
  if ( n->T->t1 == t_number ) {
    //    printf("backdate:found t1 with %ld\n",n->T->t1);
    n->T->t1 =  0;
    return(0);
  }
  if ( n->T->t2 == t_number ) {
    //printf("backdate:found t2 with %ld\n",n->T->t2);
    n->T->t2 = 0;
    return(0);
  }
  if ( n->T->t3 == t_number ) {
    // printf("backdate:found t31 with %ld\n",n->T->t3);
    n->T->t3 = 0;
    return(0);
  }
  return(-1); /* error */
}

/* deletes a triangle from tree */
void 
deleteTriangle( triag t )
{
  Node * temp;
  /* first find the neighbours of the triangle, if any */
  /* and re-adjust them to zero */
  if ( t.t1 > 0 ) {
    temp = findNode( t.t1 );
    /* now adjust neighbour, if exists  */
    if ( temp != 0 ) {
      //printf("backadating %ld for %ld\n",temp->T->n,t.n);      
      backdateNeighbour( temp, t.n );
      //printf("after backdating p(%ld %ld %ld) t(%ld %ld %ld)\n",temp->T->n1,temp->T->n2,temp->T->n3,temp->T->t1,temp->T->t2,temp->T->t3);
    }
  }
  if ( t.t2 > 0 ) {
    temp = findNode( t.t2 );
    /* now adjust neighbour, if exists  */
    if ( temp != 0 ) {
      //printf("backadating %ld for %ld\n",temp->T->n,t.n); 
      backdateNeighbour( temp, t.n );
      //printf("after backdating p(%ld %ld %ld) t(%ld %ld %ld)\n",temp->T->n1,temp->T->n2,temp->T->n3,temp->T->t1,temp->T->t2,temp->T->t3);
    }
  }
  if ( t.t3 > 0 ) {
    temp = findNode( t.t3 );
    /* now adjust neighbour, if exists  */
    if ( temp != 0 ) {
      //printf("backadating %ld for %ld\n",temp->T->n,t.n); 
      backdateNeighbour( temp, t.n );
      //printf("after backdating p(%ld %ld %ld) t(%ld %ld %ld)\n",temp->T->n1,temp->T->n2,temp->T->n3,temp->T->t1,temp->T->t2,temp->T->t3);
    }
  }
  /* now neigbours have been backdated */
  /* find the node to delete */
  temp = findNode( t.n );
  deleteNode(temp);
  /* just delete the node */
  
}

/* finds a triangle */
triag * 
findTriangle( unsigned long t_number )
{
  Node * n;
  n = findNode( t_number );
  if (n)
    return( n->T );
  return(0);
}

/* visits triangle tree */
static void 
visitTree( Node *n )
{
 
  if ((n != 0) && n != NIL) {
    visitTree( n->left );
    if (n->T != 0) {  
      printf("triangle %ld, points(%ld,%ld,%ld), neigh(%ld,%ld,%ld) bound %ld \n", n->T->n,n->T->n1, n->T->n2, n->T->n3, n->T->t1, n->T->t2, n->T->t3, n->T->boundary);
    }
    visitTree( n->right );
  }
}

/* draw the mesh on screen */ 
/* non -recursive one */
void 
visitAndDrawTree( void )
{
#ifdef NUMBER
  static char buff[10];
#endif
  triag *t;
  unsigned long int i;
 
  for ( i = 1; i <= triangle_count; i++ ) {
    t = findTriangle(i);
    if ( t != 0 ) {
      glColor3f(0.1f, 0.6f+t->epsilon, 0.4f);
      glBegin(GL_LINE_LOOP);
      /* get first point */
      glVertex2d( (GLdouble)Mx(t->n1), (GLdouble)My(t->n1) );
      /* get second point */
      glVertex2d( (GLdouble)Mx(t->n2), (GLdouble)My(t->n2) );  
      //glVertex2d( (GLdouble)p[n->T->n2].x, (GLdouble)p[n->T->n2].y );
      /* get third point */
      glVertex2d( (GLdouble)Mx(t->n3), (GLdouble)My(t->n3) );
      //glVertex2d( (GLdouble)p[n->T->n3].x, (GLdouble)p[n->T->n3].y );
      glEnd();
     
      //glFlush(); /* ? */
#ifdef NUMBER    
      /* now draw the triangle number */
      
      sprintf(buff,"%ld",t->n);
     
      /* draw at the centroid of triangle */
      glRasterPos2d((Mx(t->n1)+Mx(t->n2)+Mx(t->n3))*0.333, (My(t->n1)+My(t->n2)+My(t->n3))*0.333);
      printstring(GLUT_BITMAP_HELVETICA_10,buff);
      glFlush();
      glColor3f(0.0f, 0.3f, 1.0f);
#endif
    }
  }
}


/* non -recursive one */
void 
draw_3d( void )
{
  
 
  triag *t;
  unsigned long int i;
  long double temp;


  //  GLfloat	light_diffuse[] = { 1.0, 0.0, 0.0, 1.0 };
  //GLfloat	light_position[] = { 1.0, 1.0, 1.0, 0.0 };
  /* do something */
  //glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  //glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  //glEnable(GL_LIGHTING);
  //glEnable(GL_LIGHT0);
  //glEnable(GL_DEPTH_TEST);
  //glMatrixMode(GL_PROJECTION);
  //gluPerspective( /* field of view in degrees */ 40.0, /* aspect ratio */
  //1.0, /* Z near */ 1.0, /* Z far */ 10.0 );
  //glMatrixMode(GL_MODELVIEW);
  // gluLookAt(0.0, 0.0, 1.0, /* eye is at (0,0,5) */
  //    0.0, 0.0, 0.0, /* center is at (0,0,0) */
  //    0.0, 0.0, 0.0);  /* up is in +Y direction */
  //glTranslatef(0.0, 0.0, -1.0);
  

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
 
 /* we normalize the z coordinate before drawing */
  if ( fabsl(maxpot) > fabsl(minpot) )
    temp = fabsl(maxpot);
  else 
    temp = fabsl(minpot);
  temp = 1/temp;

  for ( i = 1; i <= triangle_count; i++ ) {
    t = findTriangle(i);
    if ( t != 0 ) {
      glBegin(GL_TRIANGLES);
      /* get first point */
      glVertex3d( (GLdouble)*((M.Narray[t->n1])->xp), (GLdouble)(M.Narray[t->n1]->y ), (GLdouble)(temp*M.Narray[t->n1]->y));
      /* get second point */
      glVertex3d( (GLdouble)*((M.Narray[t->n2])->xp), (GLdouble)(M.Narray[t->n2]->y ), (GLdouble)(temp*M.Narray[t->n2]->y));  
      //glVertex2d( (GLdouble)p[n->T->n2].x, (GLdouble)p[n->T->n2].y );
      /* get third point */
      glVertex3d( (GLdouble)*((M.Narray[t->n3])->xp), (GLdouble)(M.Narray[t->n3]->y ), (GLdouble)(temp*M.Narray[t->n3]->y));
      //glVertex2d( (GLdouble)p[n->T->n3].x, (GLdouble)p[n->T->n3].y );
      glEnd();
      //glFlush(); /* ? */

    }
  }
}

