/*  pdnmesh - a 2D finite element solver
    Copyright (c) 2001 Sarod Yatawatta <sarod@cs.pdn.ac.lk>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



/* the data types and other constants needed by all functions*/
/* first the basic data types */
/* global type definitions */
#ifndef TYPES_H
#define TYPES_H_1

typedef enum { RED, BLACK } nodecolour; /* red-black tree */
typedef enum { IN, OUT, BOUNDARY } position; /* inside or out a polygon */
typedef enum { BOUND, MESH, GRAD, ALL} option; /* menu option for plotting */
typedef enum { NU, /* neumann boundary */
	       DR } /* dirichlet boundary */ etype; /* edge type */
typedef enum { VAR, /* varible point */
	       FX /* fixed */ } value; /* point type */

/* for the point tree */
typedef struct Node_x_ {
  long double x;  /* X coordinate */
  struct Node_x_ * left;
  struct Node_x_ * right;
  struct Node_x_ * parent;

  struct Node_y_ * Ytree;
  nodecolour colour;
} node_x;

typedef struct Node_y_ {
  long double y; /* Y coordinate */
 
  struct Node_y_ * left;
  struct Node_y_ * right;
  struct Node_y_ * parent;

  unsigned long int n; /* point number */
  long double *xp; /* pointer to X coordinate */
  nodecolour colour;
   /* point data */
  long double pot; /* potential */
  position pos; /* geometric position */
  value val; /* fixed or variable */
} node_y;

typedef struct Mesh_{
  node_x * Xtree;
  node_y ** Narray;
  unsigned long int count;
} Mesh;

typedef struct point_ {
  long double x;
  long double y;
  position pos;
  /* any other problem data */
  long double pot; /* potential */
  value val;
} point;


typedef struct triag_ { /* triangle */
  unsigned long int n; /* triangle number */
  unsigned long int n1; /* point number 1*/
  unsigned long int n2; /* point number 2 */
  unsigned long int n3; /* point number 3 NOTE : in counter clockwise order*/
  /* triangle data */
  long double epsilon; /* permittivity or permiaility */
  /* to which boundary we belong ? */
  unsigned long boundary; /* number in boundary array */
//  long double rho; /* could charge density or current density */
//  rho is variable - dependent on the point - calculated at run time

  long double par; /* parameter for division */
  /* this can be assigned manually or automatically */
  /* with high par, more chance in division */

  /* now tha adjacent triangle numbers */
  /* note : t1 is opposite to n1 */
  /* t2 opposite to n2 etc..  */
  unsigned long int t1;
  unsigned long int t2;
  unsigned long int t3;
} triag;

/* now the tree of triangles */
typedef struct Node_ {
    struct Node_ *left;         /* left child */
    struct Node_ *right;        /* right child */
    struct Node_ *parent;       /* parent */
    nodecolour color;            /* node color (BLACK, RED) */
    triag *T;                     /* data stored in node */
} Node;



/* now some constants */
#define MAX 1000
#define MAXPOINTS 1000  /* initial max number of points accomodated */
#define RESIZE 10 /* number to resize the mesh by realloc() */
/* define the following for verbose output */
//#define DEBUG
/* define the following to show numbers */
//#define NUMBER
/* for 3d view */
//#define  D3VIEW



/* define sensitivity */
#define TOL 0.000000001f

/* for delaunay triangulation */
/* for frontier */
typedef struct edge_ {
  //unsigned long int n; /* number */
  unsigned long int p1; /* fist point */
  unsigned long int p2; /* next point */
  etype type; /* type of edge */
} edge;


/* data for list of edges - used in delaunay triangulation*/
typedef struct elist_ {
  edge  data;
  struct elist_* next;
} elist;



/* data structures for the parsing part */
/* node types of the syntax tree */
typedef enum { CONST, VARI, OPR } symtype;
/* symbol types */

/* constants - double */
typedef struct {
	symtype type; /* type */
	double value; /* value */
} node_const;

/* variables - lower case letters */
typedef struct {
	symtype type; /* type */
	int d; /* identifier (ASCII) of variable to point to array symval[] */
} node_var;

/* operators */
typedef struct {
	symtype type; /* type */
	int operatr;	/* type of operator */
	int nops; /* no of operands */
	union exp_nodetype_* operand[1]; /* array of operands - similar nodes */
} node_opr;

typedef union exp_nodetype_ {
	symtype type; /* either const, var or operand */
	node_const constnt;
	node_var var;
	node_opr opr;
} exp_nodetype;


/* a data structure to represent the boundaries */
typedef struct bnode_ {
  unsigned long int edge_count; /* no. of edges in this boundary */
  /* assume always that boundary is closed */
  int type; /* boundary type */
  /* 1 - normal boundary */
  /* 0 - triangles within boundary will be DELETED */
  long double epsilon; /* material properties of boundary */
  //long double rho; /* could be sigma, mu */
  /* here we use an expression for rho */
  exp_nodetype *rhoexp; /* expression for rho */
  unsigned long *edges; /* to edge array */
} bnode;

/* macros to access the mesh */
/* first to get x cordinate */
#define Mx(i)  *((M.Narray[(i)])->xp)
/* to get y coordinate */
#define My(i) (M.Narray[(i)])->y
/* to get the potential */
#define Mpot(i) (M.Narray[(i)])->pot
/* to get the position */
#define Mpos(i) (M.Narray[(i)])->pos
/* to get fixed or not */
#define Mval(i) (M.Narray[(i)])->val

/* redefine fabsl() as a macro for win32 */
#if defined(_WIN32)
#ifndef fabsl
#define 	fabsl(x)  fabs(x)
#endif /* fabsl */
#endif /* WIN32 */



/* now the functions supplied by each file */
/* 
 * pdnmesh.c
 * 
 */
/* font rendering */
extern void
printstring(void *font, char *string);

/*
 * contour.c
 *
 */
/* contour plot on screen */
extern void 
plot_contour_all(void);
/* gradient plot on screen */
extern void 
plot_gradient(void);
/* print contour on an eps file */
extern void
print_contour_all(void);

/*
 * delaunay.c
 *
 */
/* inserts the expression parsed into boundary array */
extern int
insert_expression(exp_nodetype *p, unsigned long int boundary);
/* calculate area of triangle */
extern long double
area( triag t );
/* check for three points are ccw */
extern int
ccw2( unsigned long p1, unsigned long p2, unsigned long p3);
/* removes intersections of triagulation with actual edges */
int
remove_crossings(void);
/* initial triangulation of the point set */
void
delaunay(void);

/*
 * mesh.c
 *
 */
/* initialize mesh */
void
init_mesh(Mesh *M,int maxpoints);
/* insert or check a point in mesh */
unsigned long int
insert(point p);

/*
 * refine.c
 *
 */
/* removes or modifies properties of triangles in boundaries */
extern void
cleanup_exterior(void);

/*
 * solve.c
 *
 */
/* build matrix equation and solve it*/
extern void
re_number_and_solve(void);
/* write mesh data to file */
extern void
write_mesh_file(char *file_name);
/* update splitting priorities of triangles */
extern void
update_priorities(void);

/*
 * subdivide.c
 *
 */
/* divide bad triangles */
extern int 
divide_bad_triangles(void);
/* divide triangles with three known points */
extern void
divide_corners(void);
/* refine the mesh according to delaunay criterion */
extern void
delaunay_refine(void);

/*
 * triag_tree.c
 *
 */
/* insert to tree */
extern void
insertTriangle(triag t);
/* delete from tree */
extern void
deleteTriangle(triag t);
/* finds a triangle */
extern triag *
findTriangle(unsigned long t_number);
/* draws the mesh */
extern void
visitAndDrawTree( void );
/* draws a 3D mesh */
extern void
draw_3d(void);

/*
 * lexer.l
 *
 */
/* parse equation in string */
extern int
equation_parse(char *string);

/*
 * parser.y
 *
 */
/* evaluate parse tree of expression */
extern long double
ex( exp_nodetype *p, unsigned long int pl);


/***************************************************************/
/*
 * GLOBAL VARIABLES
 *
 */
/*
 * contour.c
 *
 */
extern int contour_levels;        /* contour levels to plot */

/*
 * delaunay.c
 *
 */
extern unsigned long int npoints;/* number of points in the input, except boundary*/
extern unsigned long int nedges; 
/* total number of possible edges for used in delaunay */
/* triangulation */

extern unsigned long int bedges; /* total number of boundary edges */
extern edge *edgearray; /* edge array - including all boundaries */
extern unsigned long int nbound; /* number of boundaries */
extern bnode *boundarray; /* array of boundaries */

extern elist * frontier; /* list of edges in frontier */
extern edge hull; /* the first edge to start triangulation */
extern unsigned long int triangle_count; /* total triangles in triangulation */

extern long double xoffset; /* x offset max+min/2*/
extern long double yoffset; /* y offset max+min/2*/
extern long double xrange; /* x range max - min */
extern long double yrange; /* y range max - min */
/* using these 4, we can construct the original coordinates */
extern long double scale; /* scale */

/* global no for current boundary */
extern unsigned long int boundary_no;

/*
 * mesh.c
 *
 */
extern Mesh M; /* the mesh */
extern unsigned long int maxpoints; /* initial max limit - changed dynamically */

/*
 * pdnmesh.c
 *
 */

extern char *infile; /* input file name */

/*
 * refine.c
 *
 */

/*
 * solve.c
 *
 */
extern unsigned long int * renumber; /* re-numbering array */
extern unsigned long int * re_renumber; /* re-renumbering array */
/* used to map potentials back to the original points */

/* maximum and minimum poteltials for contours */
extern long double maxpot;
extern long double minpot;

extern long double splitter; /* to set triangle priority */

/*
 * subdivide.c
 *
 */
extern long double maximum_area; /* larges area of a triangle */
extern long double badness_limit; /* max badness for splitting */


/* 
 * triag_tree.c
 *
 */
extern Node *root; /* root of triangle tree */



#endif /* ! TYPES_H */
